from __future__ import division
import os
from .config import config
from abc import abstractmethod, ABCMeta
import numpy as np
from scipy.stats import poisson
from collections import defaultdict, OrderedDict
import tables as t
from .matrix import RegionMatrixTable, Edge, LazyEdge
from .general import MaskFilter
import msgpack
import msgpack_numpy
import math
import pandas as pd
from .tools.general import RareUpdateProgressBar, pairwise
import warnings
from future.utils import with_metaclass, viewitems
from itertools import tee
try:
    from itertools import izip as zip
except ImportError:
    pass
import logging
logger = logging.getLogger(__name__)


msgpack_numpy.patch()


try:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        os.environ["CREATE_PLOTS"] = "0"  # prevent gridmap from overriding the matplotlib backend
        gridmap_logger = logging.getLogger('gridmap.conf')
        gridmap_logger.disabled = True  # shut gridmap up about not having drmaa available
        import gridmap
        gridmap_logger.disabled = False
    has_gridmap = True

    # prepare environment for better success rate
    if 'MAX_TIME_BETWEEN_HEARTBEATS' not in os.environ:
        os.environ['MAX_TIME_BETWEEN_HEARTBEATS'] = '600'
    if 'SEND_ERROR_MAIL' not in os.environ:
        os.environ['SEND_ERROR_MAIL'] = '0'
    if 'MAX_IDLE_HEARTBEATS' not in os.environ:
        os.environ['MAX_IDLE_HEARTBEATS'] = '20'
    if 'IDLE_THRESHOLD' not in os.environ:
        os.environ['IDLE_THRESHOLD'] = '0.5'
except ImportError:
    has_gridmap = False


class PeakInfo(RegionMatrixTable):
    """
    General-purpose class for recording peaks in Hic (and similar) data.

    A peak has the following information:
    source, sink: coordinates of the highest peak pixel in the Hi-C matrix
    observed: observed value of the peak in the Hi-C matrix, generally uncorrected
    expected: expected value of the peak at this position in the Hi-C matrix
    p_value: a P-value that reflects how likely it is to observe a peak with
    these properties at random
    x, y: coordinates of the peak centroid, if it is larger than one pixel
    radius: radius of the peak, expressed in bins (can be converted to base pairs)
    """

    _classid = 'PEAKINFO'

    def __init__(self, file_name=None, mode='a', tmpdir=None,
                 _table_name_regions='regions', _table_name_peaks='edges'):
        """
        Initialize a PeakInfo object.

        :param file_name: If None, will create a working file in memory. If path to an
                          existing peak info file, will load its information. If path
                          to a non-existent file will create the file.
        :param mode: File mode, use 'a' for append, 'r' for read, and 'w' for write
        :param _table_name_regions: Internal, controls name of the region PyTables table
        :param _table_name_peaks: Internal, controls name of the peak PyTables table
        """
        additional_fields = {
            'weight': t.Float32Col(pos=2),
            'uncorrected': t.Int32Col(pos=3),
            'expected_local': t.Float32Col(pos=4),
            'oe': t.Float32Col(pos=5),
            'p_value': t.Float32Col(pos=6),
            'q_value_sum': t.Float32Col(pos=7),
            'x': t.Float32Col(pos=8),
            'y': t.Float32Col(pos=9),
            'radius': t.Float32Col(pos=10),
            'merged_pixels': t.Int32Col(pos=11),
        }

        RegionMatrixTable.__init__(self, file_name, mode=mode, tmpdir=tmpdir,
                                   additional_edge_fields=additional_fields,
                                   _table_name_regions=_table_name_regions,
                                   _table_name_edges=_table_name_peaks)

        self.peak_table = self._edges

    def _row_to_edge(self, row, lazy_edge=False, distances_in_bp=False):
        if distances_in_bp:
            bin_size = self.bin_size
        else:
            bin_size = 1

        if lazy_edge is None:
            source = row["source"]
            sink = row["sink"]
            d = dict()
            for field in self.field_names:
                if field != 'source' and field != 'sink':
                    value = row[field]
                    value = value.decode() if isinstance(value, bytes) else value
                    if field in ('x', 'y', 'radius'):
                        d[field] = value * bin_size
                    else:
                        d[field] = value

            source_node_row = self._regions[source]
            source_node = self.regions[source]
            sink_node_row = self._regions[sink]
            sink_node = self.regions[sink]
            return Peak(source_node, sink_node, **d)

        lazy_edge._row = row
        return lazy_edge

    def peaks(self, distances_in_bp=False, lazy=False):
        return self.edges(lazy=lazy, distances_in_bp=distances_in_bp)

    def _edges_subset(self, key=None, row_regions=None, col_regions=None,
                      lazy=False, lazy_edge=None, *args, **kwargs):
        if lazy:
            lazy_edge = LazyPeak(None, self._regions)
        else:
            lazy_edge = None
        return RegionMatrixTable._edges_subset(self, key=key, row_regions=row_regions,
                                               col_regions=col_regions, lazy=lazy,
                                               lazy_edge=lazy_edge, *args, **kwargs)

    def _edges_iter(self, lazy=False, lazy_edge=None, *args, **kwargs):
        if lazy:
            lazy_edge = LazyPeak(None, self._regions)
        else:
            lazy_edge = None
        return RegionMatrixTable._edges_iter(self, lazy=lazy,
                                             lazy_edge=lazy_edge,
                                             *args, **kwargs)

    def __iter__(self):
        return self.peaks()

    def filter_rao(self, queue=False):
        """
        Convenience function that applies a :class:`~RaoMergedPeakFilter`.

        :param queue: If True, filter will be queued and can be executed
                      along with other queued filters using
                      run_queued_filters
        """
        mask = self.add_mask_description('rao', 'Mask singlet peaks with a q-value sum > .02')
        rao_filter = RaoMergedPeakFilter(mask=mask)
        self.filter(rao_filter, queue)

    def to_bedpe(self, file_name, anchor_radius=True, score_field='q_value_sum', name_field=None):
        regions = list(self.regions)
        with open(file_name, 'w') as f:
            for peak in self.peaks(lazy=True, distances_in_bp=True):
                r1 = regions[peak.source]
                r2 = regions[peak.sink]
                if score_field is not None:
                    score = getattr(peak, score_field)
                else:
                    score = '.'
                if name_field is not None:
                    name = getattr(peak, name_field)
                else:
                    name = '.'

                if anchor_radius:
                    r = peak.radius
                    start1 = r1.start - r
                    end1 = r1.end + r
                    start2 = r2.start - r
                    end2 = r2.end + r
                else:
                    start1 = r1.start
                    end1 = r1.end
                    start2 = r2.start
                    end2 = r2.end

                f.write("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format(
                    r1.chromosome, int(start1), int(end1),
                    r2.chromosome, int(start2), int(end2),
                    name, score
                ))


class RaoPeakInfo(RegionMatrixTable):
    """
    Information about peaks called by :class:`~RaoPeakCaller`.

    A peak has the following information:

    source, sink: coordinates of the highest peak pixel in the Hi-C matrix
    observed: observed value of the peak in the Hi-C matrix, generally uncorrected
    e_ll: expected value of the peak given its lower-left neighborhood
    e_h: expected value of the peak given its horizontal neighborhood
    e_v: expected value of the peak given its vertical neighborhood
    e_d: expected value of the peak given its surrounding (donut) neighborhood
    e_ll_chunk: "lambda-chunk" this peak falls into given its 'll' neighborhood
    e_h_chunk: "lambda-chunk" this peak falls into given its 'h' neighborhood
    e_v_chunk: "lambda-chunk" this peak falls into given its 'v' neighborhood
    e_d_chunk: "lambda-chunk" this peak falls into given its 'd' neighborhood
    fdr_ll: FDR of the peak given its lower-left neighborhood
    fdr_h: FDR of the peak given its horizontal neighborhood
    fdr_v: FDR of the peak given its vertical neighborhood
    fdr_d: FDR of the peak given its surrounding (donut) neighborhood

    For more information about neighborhoods and peak information,
    see :class:`~RaoPeakCaller`.
    """

    _classid = 'RAOPEAKINFO'

    def __init__(self, file_name=None, mode='a', tmpdir=None,
                 _table_name_regions='regions', _table_name_peaks='edges'):
        """
        Initialize a RaoPeakInfo object.

        :param file_name: If None, will create a working file in memory. If path to an
                          existing peak info file, will load its information. If path
                          to a non-existent file will create the file.
        :param mode: File mode, use 'a' for append, 'r' for read, and 'w' for write
        :param _table_name_regions: Internal, controls name of the region PyTables table
        :param _table_name_peaks: Internal, controls name of the peak PyTables table
        """

        additional_fields = {
            'source': t.Int32Col(pos=0),
            'sink': t.Int32Col(pos=1),
            'weight': t.Float32Col(pos=2),
            'uncorrected': t.Int32Col(pos=3),
            'w': t.Int32Col(pos=4),
            'p': t.Int32Col(pos=5),
            'e_ll': t.Float32Col(pos=6, dflt=1.0),
            'e_h': t.Float32Col(pos=7, dflt=1.0),
            'e_v': t.Float32Col(pos=8, dflt=1.0),
            'e_d': t.Float32Col(pos=9, dflt=1.0),
            'oe_ll': t.Float32Col(pos=10, dflt=1.0),
            'oe_h': t.Float32Col(pos=11, dflt=1.0),
            'oe_v': t.Float32Col(pos=12, dflt=1.0),
            'oe_d': t.Float32Col(pos=13, dflt=1.0),
            'mappability_ll': t.Float32Col(pos=14, dflt=1.0),
            'mappability_h': t.Float32Col(pos=15, dflt=1.0),
            'mappability_v': t.Float32Col(pos=16, dflt=1.0),
            'mappability_d': t.Float32Col(pos=17, dflt=1.0),
            'fdr_ll': t.Float32Col(pos=18, dflt=1.0),
            'fdr_h': t.Float32Col(pos=19, dflt=1.0),
            'fdr_v': t.Float32Col(pos=20, dflt=1.0),
            'fdr_d': t.Float32Col(pos=21, dflt=1.0),
            'e_ll_chunk': t.Int32Col(pos=22, dflt=1.0),
            'e_h_chunk': t.Int32Col(pos=23, dflt=1.0),
            'e_v_chunk': t.Int32Col(pos=24, dflt=1.0),
            'e_d_chunk': t.Int32Col(pos=25, dflt=1.0),
            'll_sum': t.Int32Col(pos=26, dflt=0),
        }

        RegionMatrixTable.__init__(self, file_name, mode=mode, tmpdir=tmpdir,
                                   additional_edge_fields=additional_fields,
                                   _table_name_regions=_table_name_regions,
                                   _table_name_edges=_table_name_peaks)

        self.peak_table = self._edges

    def peaks(self, lazy=False, auto_update=True, **kwargs):
        return self.edges(lazy=lazy, auto_update=auto_update, **kwargs)

    def peaks_sorted(self, sortby, lazy=False, auto_update=True, **kwargs):
        return self.edges_sorted(sortby, lazy=lazy, auto_update=auto_update, **kwargs)

    def __iter__(self):
        return self.peaks()

    def filter_fdr(self, fdr_cutoff, queue=False):
        """
        Convenience function that applies a :class:`~FdrPeakFilter`.
        The actual algorithm and rationale used for filtering will depend on the
        internal _mapper attribute.

        :param fdr_cutoff: The false-discovery rate of every neighborhood enrichment
                           must be lower or equal to this threshold
        :param queue: If True, filter will be queued and can be executed
                      along with other queued filters using
                      run_queued_filters
        """
        mask = self.add_mask_description('fdr', 'Mask peaks with an FDR higher than %e' % fdr_cutoff)
        fdr_filter = FdrPeakFilter(fdr_cutoff=fdr_cutoff, mask=mask)
        self.filter(fdr_filter, queue)

    def filter_observed(self, cutoff, queue=False):
        """
        Convenience function that applies a :class:`~ObservedPeakFilter`.
        The actual algorithm and rationale used for filtering will depend on the
        internal _mapper attribute.

        :param cutoff: Minimum observed value
        :param queue: If True, filter will be queued and can be executed
                      along with other queued filters using
                      run_queued_filters
        """
        mask = self.add_mask_description('fdr', 'Mask peaks with an observed value lower than %e' % cutoff)
        observed_filter = ObservedPeakFilter(cutoff=cutoff, mask=mask)
        self.filter(observed_filter, queue)

    def filter_mappability(self, cutoff, queue=False):
        """
        Convenience function that applies a :class:`~MappabilityPeakFilter`.
        The actual algorithm and rationale used for filtering will depend on the
        internal _mapper attribute.

        :param cutoff: Minimum mappability (fraction of 1)
        :param queue: If True, filter will be queued and can be executed
                      along with other queued filters using
                      run_queued_filters
        """
        mask = self.add_mask_description('mappability',
                                         'Mask peaks with a mappability lower than %e' % cutoff)
        mappability_filter = MappabilityPeakFilter(mappability_cutoff=cutoff, mask=mask)
        self.filter(mappability_filter, queue)

    def filter_enrichment(self, enrichment_ll_cutoff=1.0, enrichment_h_cutoff=1.0,
                          enrichment_v_cutoff=1.0, enrichment_d_cutoff=1.0, queue=False):
        """
        Convenience function that applies a :class:`~ObservedExpectedRatioPeakFilter`.
        The actual algorithm and rationale used for filtering will depend on the
        internal _mapper attribute.

        :param queue: If True, filter will be queued and can be executed
                      along with other queued filters using
                      run_queued_filters
        """
        mask = self.add_mask_description('o/e', 'Mask peaks with a low observed/expected ratio')
        enrichment_filter = EnrichmentPeakFilter(enrichment_ll_cutoff=enrichment_ll_cutoff,
                                                 enrichment_h_cutoff=enrichment_h_cutoff,
                                                 enrichment_v_cutoff=enrichment_v_cutoff,
                                                 enrichment_d_cutoff=enrichment_d_cutoff,
                                                 mask=mask)
        self.filter(enrichment_filter, queue)

    def filter_rao(self, queue=False):
        """
        Convenience function that applies all filters Rao et al. (2014) do.

        :param queue: If True, filter will be queued and can be executed
                      along with other queued filters using
                      run_queued_filters
        """
        mask = self.add_mask_description('rao', 'Mask peaks that do not pass the RaoPeakFilter')
        rao_filter = RaoPeakFilter(mask=mask)
        self.filter(rao_filter, queue)

    @staticmethod
    def _euclidian_distance(x1, y1, x2, y2):
        """
        Determine the 2D euclidian distance between two points.
        """
        return math.sqrt((x1-x2)**2+(y1-y2)**2)

    @staticmethod
    def _centroid_and_radius(peak_list):
        """
        Determine the centroid coordinates and radius of list of peaks.
        """
        x = 0
        y = 0
        for peak in peak_list:
            x += peak.source
            y += peak.sink

        if len(peak_list) > 0:
            x /= len(peak_list)
            y /= len(peak_list)

        radius = 0
        for peak in peak_list:
            distance = RaoPeakInfo._euclidian_distance(x, y, peak.source, peak.sink)
            if distance > radius:
                radius = distance
        return x, y, radius

    def merged_peaks(self, file_name=None, euclidian_distance=20000):
        """
        Merge spatially proximal peaks.

        :param file_name: Optional file to save merged peak info to.
        :param euclidian_distance: Maximal distance in base pairs to still
                                   consider two peaks to be the same
        :return: :class:`~PeakInfo`
        """
        merged_peaks = PeakInfo(file_name=file_name, mode='w')
        merged_peaks.add_regions(self.regions(lazy=True), preserve_attributes=False)

        bin_size = self.bin_size

        def _append_merged_peak(peak_list):
            # add merged peak
            hp = peak_list[0]
            q_value_sum = hp.fdr_ll + hp.fdr_d + hp.fdr_h + hp.fdr_v
            oe = 1 if hp.e_d == 0 else hp.weight/hp.e_d
            merged_peak = Peak(source=hp.source, sink=hp.sink, weight=hp.weight,
                               uncorrected=hp.uncorrected, expected_local=hp.e_d,
                               p_value=hp.fdr_d, q_value_sum=q_value_sum, x=x, y=y,
                               radius=radius, oe=oe)
            merged_peaks.add_edge(merged_peak)

        merged_peak_counter = 0
        chromosome_names = self.chromosomes()
        for i, chromosome_name1 in enumerate(chromosome_names):
            for j in range(i, len(chromosome_names)):
                chromosome_name2 = chromosome_names[j]

                remaining_peaks_set = set()
                for peak in self.edge_subset((chromosome_name1, chromosome_name2), lazy=False):
                    remaining_peaks_set.add(peak)

                if len(remaining_peaks_set) == 0:
                    continue

                logger.info("Merging peaks in %s/%s" % (chromosome_name1, chromosome_name2))

                last_peak_number = 0
                current_peaks = []
                n_remaining = len(remaining_peaks_set)
                with RareUpdateProgressBar(max_value=n_remaining, silent=config.hide_progressbars,
                                           prefix="Peak merge") as pb:
                    while len(remaining_peaks_set) > 0:
                        x, y, radius = RaoPeakInfo._centroid_and_radius(current_peaks)

                        if len(current_peaks) == last_peak_number:
                            if len(current_peaks) > 0:
                                _append_merged_peak(current_peaks)
                                current_peaks = []
                                last_peak_number = 0
                                merged_peak_counter += 1

                            # find highest peak
                            highest_peak = None
                            for peak in remaining_peaks_set:
                                if highest_peak is None:
                                    highest_peak = peak
                                else:
                                    if highest_peak.weight < peak.weight:
                                        highest_peak = peak

                            current_peaks.append(highest_peak)
                            remaining_peaks_set.remove(highest_peak)
                        else:
                            last_peak_number = len(current_peaks)

                            closest_peak = None
                            closest_distance = None
                            for peak in remaining_peaks_set:
                                distance = RaoPeakInfo._euclidian_distance(x, y, peak.source, peak.sink)
                                if closest_peak is None or distance < closest_distance:
                                    closest_peak = peak
                                    closest_distance = distance

                            if closest_distance*bin_size <= euclidian_distance+(radius*bin_size):
                                current_peaks.append(closest_peak)
                                remaining_peaks_set.remove(closest_peak)

                        pb.update(n_remaining-len(remaining_peaks_set))

                if len(current_peaks) > 0:
                    _append_merged_peak(current_peaks)
                    merged_peak_counter += 1

        logger.info("Total merged peaks: {}".format(merged_peak_counter))
        merged_peaks.flush()
        return merged_peaks

    def parameter_tool(self, *region_pairs, **kwargs):
        from .plotting.hic_plotter import PeakParameterPlot
        import matplotlib.pyplot as plt
        parameters_plot = PeakParameterPlot(self, **kwargs)
        parameters_plot.plot(*region_pairs)
        plt.show()

        return parameters_plot.observed_cutoff, \
               parameters_plot.oe_cutoffs, \
               parameters_plot.fdr_cutoffs, \
               parameters_plot.mappability_cutoffs


class Peak(Edge):
    """
    Container for a Peak/enriched contact in a Hi-C matrix.
    """
    def __init__(self, source, sink, *args, **kwargs):
        object.__setattr__(self, 'e_ll', None)
        object.__setattr__(self, 'e_h', None)
        object.__setattr__(self, 'e_d', None)
        object.__setattr__(self, 'e_v', None)
        super(Peak, self).__init__(source, sink, *args, **kwargs)


class LazyPeak(LazyEdge):
    """
    Container for a Peak/enriched contact in a Hi-C matrix.

    This class implements :class:`~LazyPeak`, which provides lazy
    loading of attributes from a PyTables table row.
    """
    def __init__(self, row, nodes_table, bin_size=1):
        LazyEdge.__init__(self, row, nodes_table)
        self._bin_size = bin_size

    def __getattr__(self, item):
        res = LazyEdge.__getattr__(self, item)
        if item in ('x', 'y', 'radius'):
            return self._bin_size * res
        return res


class PeakFilter(with_metaclass(ABCMeta, MaskFilter)):
    """
    Abstract class that provides filtering functionality for the
    peaks in a :class:`~RaoPeakInfo` object.

    Extends MaskFilter and overrides valid(self, row) to make
    :class:`~RaoPeakInfo` filtering more "natural".

    To create custom filters for the :class:`~RapPeakInfo` object, extend this
    class and override the valid_peak(self, peak) method.
    valid_peak should return False for a specific :class:`~Edge` object
    if the object is supposed to be filtered/masked and True
    otherwise. See :class:`~DiagonalFilter` for an example.

    Pass a custom filter to the :func:`~RaoPeakInfo.filter` method in :class:`~Hic`
    to apply it.
    """

    def __init__(self, mask=None):
        """
        Initialize PeakFilter.

        :param mask: The Mask object that should be used to mask
                     filtered :class:`~RaoPeak` objects. If None the default
                     Mask will be used.
        """
        super(PeakFilter, self).__init__(mask)

    @abstractmethod
    def valid_peak(self, peak):
        """
        Determine if a :class:`~RaoPeak` object is valid or should
        be filtered.

        When implementing custom PeakFilter this method must be
        overridden. It should return False for :class:`~RaoPeak` objects that
        are to be fitered and True otherwise.

        Internally, the :class:`~RaoPeakInfo` object will iterate over all RaoPeak
        instances to determine their validity on an individual
        basis.

        :param peak: A :class:`~RaoPeak` object
        :return: True if :class:`~PeakFilter` is valid, False otherwise
        """
        pass

    def valid(self, row):
        """
        Map valid_peak to MaskFilter.valid(self, row).

        :param row: A pytables Table row.
        :return: The boolean value returned by valid_edge.
        """
        peak = LazyEdge(row)
        return self.valid_peak(peak)


class ObservedPeakFilter(PeakFilter):
    """
    Filter for peaks that do not pass a certain FDR cutoff.
    """
    def __init__(self, cutoff=1, mask=None):
        """
        Initialize filter object.

        :param mask: A :class:`~fanc.data.general.Mask` object.
        :param cutoff: Minimum observed value to consider peak
        """
        super(ObservedPeakFilter, self).__init__(mask=mask)
        self.cutoff = cutoff

    def valid_peak(self, peak):
        """
        Evaluate whether a peak passes FDR cutoffs set in __init__
        :param peak: An :class:`~fanc.data.genomic.Edge` object
        :return: True if peak passes interal FDR cutoffs, False otherwise
        """
        if peak.uncorrected < self.cutoff:
            return False
        return True


class DistancePeakFilter(PeakFilter):
    """
    Filter for peaks where regions are closer than this cutoff in bins.
    """
    def __init__(self, cutoff=1, mask=None):
        """
        Initialize filter object.

        :param mask: A :class:`~fanc.data.general.Mask` object.
        :param cutoff: Minimum observed value to consider peak
        """
        super(DistancePeakFilter, self).__init__(mask=mask)
        self.cutoff = cutoff

    def valid_peak(self, peak):
        """
        Evaluate whether a peak passes FDR cutoffs set in __init__
        :param peak: An :class:`~fanc.data.genomic.Edge` object
        :return: True if peak passes internal FDR cutoffs, False otherwise
        """
        if abs(peak.source - peak.sink) < self.cutoff:
            return False
        return True


class FdrPeakFilter(PeakFilter):
    """
    Filter for peaks that do not pass a certain FDR cutoff.
    """
    def __init__(self, mask=None, fdr_cutoff=None,
                 fdr_ll_cutoff=None, fdr_v_cutoff=None,
                 fdr_h_cutoff=None, fdr_d_cutoff=None):
        """
        Initialize filter object.

        :param mask: A :class:`~fanc.data.general.Mask` object.
        :param fdr_cutoff: Global FDR cutoff. Is overridden by the
                           neighborhood-specific cutoffs
        :param fdr_ll_cutoff: FDR cutoff for the lower-left neighborhood
        :param fdr_v_cutoff: FDR cutoff for the vertical neighborhood
        :param fdr_h_cutoff: FDR cutoff for the horizontal neighborhood
        :param fdr_d_cutoff: FDR cutoff for the donut neighborhood
        """
        super(FdrPeakFilter, self).__init__(mask=mask)
        self.fdr_ll_cutoff = fdr_cutoff
        self.fdr_h_cutoff = fdr_cutoff
        self.fdr_v_cutoff = fdr_cutoff
        self.fdr_d_cutoff = fdr_cutoff
        if fdr_ll_cutoff is not None:
            self.fdr_ll_cutoff = fdr_ll_cutoff
        if fdr_h_cutoff is not None:
            self.fdr_h_cutoff = fdr_h_cutoff
        if fdr_v_cutoff is not None:
            self.fdr_v_cutoff = fdr_v_cutoff
        if fdr_d_cutoff is not None:
            self.fdr_d_cutoff = fdr_d_cutoff

    def valid_peak(self, peak):
        """
        Evaluate whether a peak passes FDR cutoffs set in __init__
        :param peak: An :class:`~fanc.data.genomic.Edge` object
        :return: True if peak passes interal FDR cutoffs, False otherwise
        """
        if self.fdr_ll_cutoff is not None and peak.fdr_ll > self.fdr_ll_cutoff:
            return False
        if self.fdr_h_cutoff is not None and peak.fdr_h > self.fdr_h_cutoff:
            return False
        if self.fdr_v_cutoff is not None and peak.fdr_v > self.fdr_v_cutoff:
            return False
        if self.fdr_d_cutoff is not None and peak.fdr_d > self.fdr_d_cutoff:
            return False
        return True


class MappabilityPeakFilter(PeakFilter):
    """
    Filter for peaks that do not pass a certain FDR cutoff.
    """
    def __init__(self, mask=None, mappability_cutoff=None,
                 mappability_ll_cutoff=None, mappability_v_cutoff=None,
                 mappability_h_cutoff=None, mappability_d_cutoff=None):
        """
        Initialize filter object.

        :param mask: A :class:`~fanc.data.general.Mask` object.
        :param mappability_cutoff: Global mappability cutoff. Is overridden by the
                                   neighborhood-specific cutoffs
        :param mappability_ll_cutoff: Mappability cutoff for the lower-left neighborhood
        :param mappability_v_cutoff: Mappability cutoff for the vertical neighborhood
        :param mappability_h_cutoff: Mappability cutoff for the horizontal neighborhood
        :param mappability_d_cutoff: Mappability cutoff for the donut neighborhood
        """
        super(MappabilityPeakFilter, self).__init__(mask=mask)
        self.mappability_ll_cutoff = mappability_cutoff
        self.mappability_h_cutoff = mappability_cutoff
        self.mappability_v_cutoff = mappability_cutoff
        self.mappability_d_cutoff = mappability_cutoff
        if mappability_ll_cutoff is not None:
            self.mappability_ll_cutoff = mappability_ll_cutoff
        if mappability_h_cutoff is not None:
            self.mappability_h_cutoff = mappability_h_cutoff
        if mappability_v_cutoff is not None:
            self.mappability_v_cutoff = mappability_v_cutoff
        if mappability_d_cutoff is not None:
            self.mappability_d_cutoff = mappability_d_cutoff

    def valid_peak(self, peak):
        """
        Evaluate whether a peak passes FDR cutoffs set in __init__
        :param peak: An :class:`~fanc.data.genomic.Edge` object
        :return: True if peak passes interal FDR cutoffs, False otherwise
        """
        if self.mappability_ll_cutoff is not None and peak.mappability_ll < self.mappability_ll_cutoff:
            return False
        if self.mappability_h_cutoff is not None and peak.mappability_h < self.mappability_h_cutoff:
            return False
        if self.mappability_v_cutoff is not None and peak.mappability_v < self.mappability_v_cutoff:
            return False
        if self.mappability_d_cutoff is not None and peak.mappability_d < self.mappability_d_cutoff:
            return False
        return True


class EnrichmentPeakFilter(PeakFilter):
    """
    Filter peaks that do not have a sufficiently strong observed/expected ratio.
    """
    def __init__(self, enrichment_cutoff=None, 
                 enrichment_ll_cutoff=None, enrichment_h_cutoff=None,
                 enrichment_v_cutoff=None, enrichment_d_cutoff=None, 
                 mask=None):
        """
        Initialize filter object.

        :param ll_ratio: Minimum observed/e_ll ratio
        :param h_ratio: Minimum observed/e_h ratio
        :param v_ratio: Minimum observed/e_v ratio
        :param d_ratio: Minimum observed/e_d ratio
        :param mask: A :class:`~fanc.data.general.Mask` object.
        :return: True if all observed/expected ratios pass the thresholds,
                 False otherwise
        """
        super(EnrichmentPeakFilter, self).__init__(mask=mask)
        self.enrichment_ll_cutoff = enrichment_cutoff
        self.enrichment_h_cutoff = enrichment_cutoff
        self.enrichment_v_cutoff = enrichment_cutoff
        self.enrichment_d_cutoff = enrichment_cutoff
        if enrichment_ll_cutoff is not None:
            self.enrichment_ll_cutoff = enrichment_ll_cutoff
        if enrichment_h_cutoff is not None:
            self.enrichment_h_cutoff = enrichment_h_cutoff
        if enrichment_v_cutoff is not None:
            self.enrichment_v_cutoff = enrichment_v_cutoff
        if enrichment_d_cutoff is not None:
            self.enrichment_d_cutoff = enrichment_d_cutoff

    def valid_peak(self, peak):
        # covering my ass for legacy_old bug
        if peak.e_d == 0 or peak.e_ll == 0 or peak.e_h == 0 or peak.e_v == 0:
            return False

        if self.enrichment_ll_cutoff is not None and peak.oe_ll < self.enrichment_ll_cutoff:
            return False
        if self.enrichment_h_cutoff is not None and peak.oe_h < self.enrichment_h_cutoff:
            return False
        if self.enrichment_v_cutoff is not None and peak.oe_v < self.enrichment_v_cutoff:
            return False
        if self.enrichment_d_cutoff is not None and peak.oe_d < self.enrichment_d_cutoff:
            return False
        return True


class RaoPeakFilter(PeakFilter):
    """
    Filter peaks exactly the same way that Rao et al. (2014) do.

    It only retains peaks that

    1. are at least 2-fold enriched over either the donut or lower-left neighborhood
    2. are at least 1.5-fold enriched over the horizontal and vertical neighborhoods
    3. are at least 1.75-fold enriched over both the donut and lower-left neighborhood
    4. have an FDR <= 0.1 in every neighborhood
    """
    def __init__(self, mask=None):
        PeakFilter.__init__(self, mask=mask)

    def valid_peak(self, peak):
        # covering my ass for legacy_old bug
        if peak.e_d == 0 or peak.e_ll == 0 or peak.e_h == 0 or peak.e_v == 0:
            return False

        # 1.
        if peak.oe_d <= peak.oe_ll < 2.0:
            return False

        # 2.
        if peak.oe_h < 1.5 and peak.oe_v < 1.5:
            return False

        # 3.
        if peak.oe_d < 1.75 or peak.oe_ll < 1.75:
            return False

        # 4.
        if peak.fdr_d > .1:
            return False
        if peak.fdr_ll > .1:
            return False
        if peak.fdr_h > .1:
            return False
        if peak.fdr_v > .1:
            return False

        return True


class RaoMergedPeakFilter(PeakFilter):
    """
    Filter merged peaks exactly the same way that Rao et al. (2014) do.

    It removes peaks that are singlets and have a q-value sum >.02.
    """
    def __init__(self, cutoff=.02, mask=None):
        PeakFilter.__init__(self, mask=mask)
        self.cutoff = cutoff

    def valid_peak(self, peak):
        if peak.radius == 0 and peak.q_value_sum > self.cutoff:
            return False

        return True


class FdrSumFilter(PeakFilter):
    """
    Remove peaks that have a q-value sum > cutoff.
    """
    def __init__(self, cutoff=1.0, mask=None):
        PeakFilter.__init__(self, mask=mask)
        self.cutoff = cutoff

    def valid_peak(self, peak):
        if peak.q_value_sum > self.cutoff:
            return False

        return True


class RaoPeakCaller(object):
    """
    Class that calls peaks the same way Rao et al. (2014) propose.

    Every pixel in a Hi-C matrix is evaluated based on its local
    "neighborhood", i.e. the pixel's observed value is compared
    to the expected value calculated from all pixels in its close
    surroundings.

    If a pixel is significantly enriched with respect to all
    investigated neighborhoods, it is assumed to be a "peak".

    Four neighborhood types are calculated:

    - "donut": Pixels surrounding the investigated pixel in a
      certain distance range
    - "lower-left" Pixels to the "lower-left" of a given pixel
    - "horizontal": Pixels left and right of a given pixel
    - "vertical": Pixels above and below a given pixel

    While the first neighborhood type most generally calculates
    enrichment of local background, the other types of
    neighborhoods serve mostly to exclude false-positive results
    from non-peak structures, such as TAD boundaries.

    The :class:`~RaoPeakCaller` is initialized with the peak
    calling parameters and run using :func:`~RaoPeakCaller.call_peaks`.

    FDRs for intra-chromosomal peaks are automatically corrected for multiple
    testing using the "lamda-chunking" methodology introduced in Rao et al. 2014.
    FDRs for inter-chromosomal peaks are corrected by default using the Benjamini
    Hochberg false-discovery rate correction (but 'bonferroni' is also an option)
    """

    def __init__(self, p=None, w_init=None, min_locus_dist=None, max_w=20, min_ll_reads=16,
                 process_inter=False, correct_inter='fdr', n_processes=4,
                 slice_size=200, min_mappable_fraction=0.7, cluster=False):
        """
        Initialize RaoPeakCaller with peak calling parameters.

        :param p: (int) "padding" of pixels around invesitgated peak
        :param w_init: initial width of the area around a pixel to investigate
        :param min_locus_dist: Minimal distance in bins between two loci to consider peak
        :param max_w: Maximal width after extending investigated area around peak
        :param min_ll_reads: Threshold for the number of reads in the lower-left
                             neighborhood of a pixel to consider it as a peak
        :param process_inter: If False, ignores inter-chromosomal peaks
        :param correct_inter: If None, does not correct inter-chromosomal peaks for
                              multiple testing. Other options are 'fdr' for Benjamini-
                              Hochberg correction of 'bonferroni' for Bonferroni correction
        :param n_processes: Number of processes to use for peak calling.
        :param slice_size: length of the matrix square investigated by each process.
        :param cluster: If True, attempts to call peaks using an SGE cluster. If False,
                        will use multiprocessing.
        """
        self.p = p
        self.w_init = w_init
        if p is not None and w_init is not None:
            if not w_init > p:
                raise ValueError("w_init ({}) must be strictly greater than p ({})!".format(w_init, p))

        self.min_locus_dist = min_locus_dist
        self.max_w = max_w
        self.min_ll_reads = min_ll_reads
        self.process_inter = process_inter
        self.correct_inter = correct_inter
        self.n_processes = n_processes
        self.slice_size = slice_size
        self.min_mappable_fraction = min_mappable_fraction
        self.cluster = cluster
        if self.cluster is True:
            if not has_gridmap:
                logger.warning("Cannot use the cluster because of previous error.")
                self.cluster = False

        super(RaoPeakCaller, self).__init__()

    # sum of reads in lower-left neighborhood
    @staticmethod
    def ll_sum(m, i, j, w=1, p=0):
        """
        Compute the sum of pixels in the lower-left neighborhood of a pixel.
        """
        i_max, j_max = m.shape

        sum1 = np.ma.sum(m[max(0, i+1):min(i_max, i+w+1), max(0, j-w):min(j_max, j)])
        sum2 = np.ma.sum(m[max(0, i+1):min(i_max, i+p+1), max(0, j-p):min(j_max, j)])

        return sum1 - sum2

    @staticmethod
    def enrichment(m, e, i, j, neighborhood_method, w=1, p=0):
        dividend = neighborhood_method(m, i, j, w, p)
        divisor = neighborhood_method(e, i, j, w, p)
        return float(dividend / divisor * e[i, j])

    @staticmethod
    def e_ll_sum(m, i, j, w=1, p=0):
        i_max, j_max = m.shape
        sum1 = np.ma.sum(m[max(0, i + 1):min(i + w + 1, i_max), max(0, j - w):min(j_max, j)])
        sum2 = np.ma.sum(m[max(0, i + 1):min(i_max, i + p + 1), max(0, j - p):min(j_max, j)])
        return sum1 - sum2

    @staticmethod
    def e_h_sum(m, i, j, w=1, p=0):
        i_max, j_max = m.shape
        sum1 = np.ma.sum(m[max(0, i - 1):min(i_max, i + 2), max(0, j - w):min(j_max, j + w + 1)])
        sum2 = np.ma.sum(m[max(0, i - 1):min(i_max, i + 2), max(0, j - p):min(j_max, j + p + 1)])
        return sum1 - sum2

    @staticmethod
    def e_v_sum(m, i, j, w=1, p=0):
        i_max, j_max = m.shape
        sum1 = np.ma.sum(m[max(0, i - w):min(i_max, i + w + 1), max(0, j - 1):min(j_max, j + 2)])
        sum2 = np.ma.sum(m[max(0, i - p):min(i_max, i + p + 1), max(0, j - 1):min(j_max, j + 2)])
        return sum1 - sum2

    @staticmethod
    def e_d_sum(m, i, j, w=1, p=0):
        i_max, j_max = m.shape
        sum1 = np.ma.sum(m[max(0, i - w):min(i_max, i + w + 1), max(0, j - w):min(j_max, j + w + 1)])
        sum2 = np.ma.sum(m[max(0, i - p):min(i_max, i + p + 1), max(0, j - p):min(j_max, j + p + 1)])
        sum3 = np.ma.sum(m[max(0, i - w):max(0, i - p), j])
        sum4 = np.ma.sum(m[max(0, i + p + 1):min(i_max, i + w + 1), j])
        sum5 = np.ma.sum(m[i, max(0, j - w):max(0, j - p)])
        sum6 = np.ma.sum(m[i, max(0, j + p + 1):min(j_max, j + w + 1)])

        sum1 = 0 if np.ma.is_masked(sum1) else sum1
        sum2 = 0 if np.ma.is_masked(sum2) else sum2
        sum3 = 0 if np.ma.is_masked(sum3) else sum3
        sum4 = 0 if np.ma.is_masked(sum4) else sum4
        sum5 = 0 if np.ma.is_masked(sum5) else sum5
        sum6 = 0 if np.ma.is_masked(sum6) else sum6

        return sum1 - sum2 - sum3 - sum4 - sum5 - sum6

    # lower-left neighborhood
    @staticmethod
    def e_ll(m, i, j, e, w=1, p=0):
        """
        Compute the average value of pixels in the lower-left neighborhood of a pixel.
        """
        return RaoPeakCaller.enrichment(m, e, i, j, RaoPeakCaller.e_ll_sum, w=w, p=p)

    # horizontal neighborhood
    @staticmethod
    def e_h(m, i, j, e, w=1, p=0):
        """
        Compute the average value of pixels in the horizontal neighborhood of a pixel.
        """
        return RaoPeakCaller.enrichment(m, e, i, j, RaoPeakCaller.e_h_sum, w=w, p=p)

    # vertical neighborhood
    @staticmethod
    def e_v(m, i, j, e, w=1, p=0):
        """
        Compute the average value of pixels in the vertical neighborhood of a pixel.
        """
        return RaoPeakCaller.enrichment(m, e, i, j, RaoPeakCaller.e_v_sum, w=w, p=p)

    # donut neighborhood
    @staticmethod
    def e_d(m, i, j, e, w=1, p=0):
        """
        Compute the average value of pixels in the "donut" neighborhood of a pixel.
        """
        return RaoPeakCaller.enrichment(m, e, i, j, RaoPeakCaller.e_d_sum, w=w, p=p)

    @staticmethod
    def find_chunk(value, chunk_func=lambda x: 3*np.log2(x)):
        """
        Use bisection to find a matching lambda chunk for a given expected value.
        """
        if value is None or np.isnan(value):
            return None
        if value < 1:
            return 0
        v = chunk_func(value)
        if not np.isfinite(v):
            return None
        return max(0, int(v) + 1)

    def _process_jobs(self, jobs, peaks, observed_chunk_distribution):
        """
        Process the output from :func:`~process_matrix_range` and save in peak table.
        """
        # if the grid does not work for some reason, this will fall back on
        # multiprocessing itself
        logger.debug("Getting gridmap output...")

        job_kwargs = {}
        if config.gridmap_tmpdir is not None:
            job_kwargs['temp_dir'] = config.gridmap_tmpdir
        job_outputs = gridmap.process_jobs(jobs, max_processes=self.n_processes,
                                           local=not self.cluster, **job_kwargs)
        logger.debug("Got gridmap output.")

        for compressed_results in job_outputs:
            results = msgpack.loads(compressed_results, strict_map_key=False)
            for result in results:
                found_none = False
                for value in result:
                    if value is None:
                        found_none = True
                if found_none:
                    continue

                source, sink, weight, w_corr, p, observed, \
                    ll_sum, e_ll, e_v, e_h, e_d, \
                    o_chunk, e_ll_chunk, e_v_chunk, e_h_chunk, e_d_chunk, \
                    e_ll_mappable, e_v_mappable, e_h_mappable, e_d_mappable = result

                # update observed distribution
                observed_chunk_distribution['ll'][e_ll_chunk][observed] += 1
                observed_chunk_distribution['h'][e_h_chunk][observed] += 1
                observed_chunk_distribution['v'][e_v_chunk][observed] += 1
                observed_chunk_distribution['d'][e_d_chunk][observed] += 1

                if weight == 0.0:
                    continue

                oe_ll = 1 if e_ll == 0 else weight / e_ll
                oe_h = 1 if e_h == 0 else weight / e_h
                oe_v = 1 if e_v == 0 else weight / e_v
                oe_d = 1 if e_d == 0 else weight / e_d

                peak = Edge(source=source, sink=sink,
                            weight=weight, uncorrected=observed,
                            w=w_corr, p=p, ll_sum=ll_sum,
                            e_ll=e_ll, e_h=e_h, e_v=e_v, e_d=e_d,
                            oe_ll=oe_ll, oe_h=oe_h, oe_v=oe_v, oe_d=oe_d,
                            e_ll_chunk=e_ll_chunk, e_v_chunk=e_v_chunk,
                            e_h_chunk=e_h_chunk, e_d_chunk=e_d_chunk,
                            mappability_ll=e_ll_mappable, mappability_v=e_v_mappable,
                            mappability_h=e_h_mappable, mappability_d=e_d_mappable)
                peaks.add_edge(peak)

    @staticmethod
    def _get_fdr_cutoffs(observed_chunk_distribution, e_func=lambda x: 2**(x/3)):
        """
        For all possible observed values in each lambda chunk, determine the
        FDR cutoff that denotes the lower significance bound.
        """
        # determine all possible observed values

        fdr_cutoffs = dict()
        for e_type in observed_chunk_distribution:  # ll, h, v, d
            fdr_cutoffs[e_type] = defaultdict(lambda: defaultdict(int))
            for chunk in observed_chunk_distribution[e_type].keys():
                max_e = e_func(chunk)
                poisson_e = poisson(max_e)

                observed_sum = 0
                for observed_count in observed_chunk_distribution[e_type][chunk].values():
                    observed_sum += observed_count

                observed_distribution_integral_left = 0
                for observed in sorted(observed_chunk_distribution[e_type][chunk].keys()):
                    observed_count = observed_chunk_distribution[e_type][chunk][observed]
                    observed_distribution_integral_left += observed_count/observed_sum
                    observed_distribution_integral_right = 1 - observed_distribution_integral_left
                    poisson_integral_right = 1 - poisson_e.cdf(observed)

                    if observed_distribution_integral_right > 0:
                        integral_ratio = poisson_integral_right/observed_distribution_integral_right
                        fdr_cutoff = min(1, integral_ratio)
                        fdr_cutoffs[e_type][chunk][observed] = fdr_cutoff
                    else:
                        fdr_cutoffs[e_type][chunk][observed] = 0
        return fdr_cutoffs

    @staticmethod
    def segment_matrix_intra(m, chunk_size, w_max):
        for i in range(0, m.shape[0], chunk_size):
            i_start = max(0, i - w_max)
            i_end = min(i + chunk_size + w_max, m.shape[0])
            i_range = (i_start, i_end)
            i_inspect = (i, min(i + chunk_size, m.shape[0]))
            for j in range(i, m.shape[1], chunk_size):
                j_start = max(0, j - w_max)
                j_end = min(j + chunk_size + w_max, m.shape[1])
                j_range = (j_start, j_end)
                j_inspect = (j, min(j + chunk_size, m.shape[1]))
                ms = m[i_start:i_end, j_start:j_end]
                yield ms, i_range, i_inspect, j_range, j_inspect

    def _find_peaks_intra_matrix(self, m, e, c, peak_info, mappable, ix_offset,
                                 observed_chunk_distribution, w, p):
        """
        Given a matrix (strictly intra-chromosomal), calculate peak
        information for all pixels.
        """

        jobs = []
        for segment in RaoPeakCaller.segment_matrix_intra(m, self.slice_size, self.max_w):
            ms, i_range, i_inspect, j_range, j_inspect = segment

            args = [ms, e, ix_offset,
                    i_range, i_inspect, mappable[i_range[0]:i_range[1]], c[i_range[0]:i_range[1]],
                    j_range, j_inspect, mappable[j_range[0]:j_range[1]], c[j_range[0]:j_range[1]],
                    w, p, self.min_locus_dist, self.min_ll_reads, self.min_mappable_fraction,
                    self.max_w]

            args = msgpack.dumps(args)
            job = gridmap.Job(process_matrix_segment_intra, [args])
            jobs.append(job)

            # submit intermediate segments if maximum number of jobs reached
            if len(jobs) >= self.n_processes:
                self._process_jobs(jobs, peak_info, observed_chunk_distribution)
                jobs = []

        if len(jobs) > 0:
            self._process_jobs(jobs, peak_info, observed_chunk_distribution)

    def call_peaks(self, hic, chromosome_pairs=None, file_name=None, intra_expected=None, inter_expected=None):
        """
        Call peaks in Hi-C matrix.

        This method will determine each pixel's likelihood to
        be a "true" peak. By default, only pixels with non-zero count and
        an observed/expected ratio > 1.0 for each neighborhood will be
        reported, because these can by definition not be true peaks.

        The peak calling behavior can be influenced by modifying
        the object attributes set when initializing :class:`~RaoPeakCaller`.

        :param hic: A :class:`~fanc.data.genomic.Hic` object
        :param chromosome_pairs: If None, all chromosome pairs will be
                                 investigated for peaks. Otherwise
                                 specify a list of chromosome name
                                 tuples (e.g. [('chr1', 'chr1'),
                                 ('chr1', 'chr3'), ...])
        :param file_name: An optional filename that backs the returned
                          :class:`~RaoPeakInfo` object
        :param intra_expected: A dict of the form
                               <chromosome>:<list of expected values> to override
                               expected value calculation
        :param inter_expected: A float describing the expected value
                               for inter-chromosomal contact matrix entries
        :return: :class:`~RaoPeakInfo` object
        """
        if self.process_inter:
            raise RuntimeError("Inter-chromosomal peak calling not currently supported!")

        peaks = RaoPeakInfo(file_name, mode='w')
        peaks.add_regions(hic.regions, preserve_attributes=False)

        # expected values
        if intra_expected is None:
            logger.info("Calculating intra-chromosomal expected values...")
            _, intra_expected, _ = hic.expected_values()
            logger.info("Done.")
        # if self.process_inter and inter_expected is None:
        #     logger.info("Inter-chromosomal expected values...")
        #     with ExpectedContacts(hic, smooth=False) as expected:
        #         inter_expected = expected.inter_expected()
        #
        # if self.process_inter:
        #     intra_possible, inter_possible = hic.possible_contacts()
        # else:
        #     intra_possible, inter_possible = None, None

        # mappability
        mappable = hic.mappable()
        
        # initialize peak parameters
        p = self.p
        w_init = self.w_init

        # empirically choose p and w_init
        # if not already chosen
        if p is None or w_init is None:
            bin_size = hic.bin_size
            if bin_size > 25000:
                p = 1 if p is None else p
                w_init = 3 if w_init is None else w_init
            else:
                p = int(24999/bin_size) if p is None else p
                w_init = int(25000/bin_size + 0.5) + 2 if w_init is None else w_init
        if self.min_locus_dist is None:
            self.min_locus_dist = p
        logger.info("Initial parameter values: p=%d, w=%d" % (p, w_init))

        logger.info("Obtaining bias vector...")
        c = hic.bias_vector()
        logger.info("Done.")

        # lambda chunks container
        observed_chunk_distribution = dict()
        for e_type in ('ll', 'h', 'v', 'd'):
            observed_chunk_distribution[e_type] = defaultdict(lambda: defaultdict(int))

        # start processing chromosome pairs
        if chromosome_pairs is None:
            chromosome_pairs = []
            chromosomes = list(hic.chromosomes())

            for i_chr, chromosome1 in enumerate(chromosomes):
                for j_chr in range(i_chr, len(chromosomes)):
                    chromosome2 = chromosomes[j_chr]
                    chromosome_pairs.append((chromosome1, chromosome2))

        chromosome_bins = hic.chromosome_bins
        for chromosome1, chromosome2 in chromosome_pairs:
            logger.info("Processing %s-%s" % (chromosome1, chromosome2))

            start1, end1 = chromosome_bins[chromosome1]
            ix_offset = start1
            start2, end2 = chromosome_bins[chromosome2]
            if chromosome1 == chromosome2:
                m = hic.matrix((chromosome1, chromosome2))
                self._find_peaks_intra_matrix(m, intra_expected[chromosome1], c[start1:end1],
                                              peaks, mappable[start1:end1], ix_offset,
                                              observed_chunk_distribution, w_init, p)
            elif self.process_inter:
                warnings.warn("Inter-chromosomal peak calling not currently supported!")
                # self._find_peaks_inter_matrix(m, inter_expected, c[start1:end1], c[start2:end2],
                #                              peaks, mappable[start1:end1], mappable[start2:end2],
                #                              None, None, w_init, p)
        peaks.flush()

        # calculate fdrs
        logger.info("Finding FDR cutoffs...")
        fdr_cutoffs = RaoPeakCaller._get_fdr_cutoffs(observed_chunk_distribution)

        # regions_dict = peaks.regions_dict
        with RareUpdateProgressBar(max_value=len(peaks.edges), prefix='FDR') as pb:
            for i, peak in enumerate(peaks.peaks(lazy=True, writable=True)):
                # region1 = regions_dict[peak.source]
                # region2 = regions_dict[peak.sink]
                # if region1.chromosome == region2.chromosome:
                try:
                    peak.fdr_ll = fdr_cutoffs['ll'][peak.e_ll_chunk][peak.uncorrected]
                    peak.fdr_h = fdr_cutoffs['h'][peak.e_h_chunk][peak.uncorrected]
                    peak.fdr_v = fdr_cutoffs['v'][peak.e_v_chunk][peak.uncorrected]
                    peak.fdr_d = fdr_cutoffs['d'][peak.e_d_chunk][peak.uncorrected]
                except KeyError:
                    peak.fdr_ll = 1
                    peak.fdr_h = 1
                    peak.fdr_v = 1
                    peak.fdr_d = 1
                peak.update()

                pb.update(i)
                # else:
                #     # Bonferroni correction
                #     if self.correct_inter == 'bonferroni':
                #         peak.fdr_ll *= inter_possible
                #         peak.fdr_h *= inter_possible
                #         peak.fdr_v *= inter_possible
                #         peak.fdr_d *= inter_possible
                #         peak.update()
        peaks.flush()

        # if self.process_inter and self.correct_inter == 'fdr':
        #     # fdr_ll
        #     for i, peak in enumerate(peaks.peaks_sorted('fdr_ll', lazy=True, auto_update=False)):
        #         region1 = regions_dict[peak.source]
        #         region2 = regions_dict[peak.sink]
        #         if region1.chromosome != region2.chromosome:
        #             peak.fdr_ll *= inter_possible/(i+1)
        #             peak.update()
        #     peaks.flush()
        #     # fdr_h
        #     for i, peak in enumerate(peaks.peaks_sorted('fdr_h', lazy=True, auto_update=False)):
        #         region1 = regions_dict[peak.source]
        #         region2 = regions_dict[peak.sink]
        #         if region1.chromosome != region2.chromosome:
        #             peak.fdr_h *= inter_possible/(i+1)
        #             peak.update()
        #     peaks.flush()
        #     # fdr_v
        #     for i, peak in enumerate(peaks.peaks_sorted('fdr_v', lazy=True, auto_update=False)):
        #         region1 = regions_dict[peak.source]
        #         region2 = regions_dict[peak.sink]
        #         if region1.chromosome != region2.chromosome:
        #             peak.fdr_v *= inter_possible/(i+1)
        #             peak.update()
        #     peaks.flush()
        #     # fdr_d
        #     for i, peak in enumerate(peaks.peaks_sorted('fdr_d', lazy=True, auto_update=False)):
        #         region1 = regions_dict[peak.source]
        #         region2 = regions_dict[peak.sink]
        #         if region1.chromosome != region2.chromosome:
        #             peak.fdr_d *= inter_possible/(i+1)
        #             peak.update()
        #     peaks.flush()

        return peaks


def process_matrix_segment_intra(data):
    m_original, e, ix_offset, \
        i_range, i_inspect, mappable_i, c_i, \
        j_range, j_inspect, mappable_j, c_j, \
        w, p, min_locus_dist, min_ll_reads, min_mappable, \
        max_w = msgpack.loads(data, strict_map_key=False)

    m_original = np.array(m_original)
    # construct convenient matrices
    row_ixs = np.arange(i_range[0], i_range[1])
    col_ixs = np.arange(j_range[0], j_range[1])
    with np.errstate(divide='ignore', invalid='ignore'):
        m_uncorrected = np.rint(m_original/c_i[:, None]/c_j)
    m_distance = np.array([abs(col_ixs - i) for i in row_ixs])
    expected_f = np.vectorize(lambda x: e[x])
    m_expected = expected_f(m_distance)

    # mask above matrices by mappability
    mask = np.zeros(m_original.shape, dtype=bool)
    mask[np.logical_not(mappable_i)] = True
    mask[:, np.logical_not(mappable_j)] = True

    m_original = np.ma.masked_where(mask, m_original)
    m_uncorrected = np.ma.masked_where(mask, m_uncorrected)
    m_expected = np.ma.masked_where(mask, m_expected)
    m_ones = np.ones(m_original.shape)

    results = []
    for o_i in range(i_inspect[0], i_inspect[1]):
        i = o_i - i_range[0]
        for o_j in range(j_inspect[0], j_inspect[1]):
            j = o_j - j_range[0]

            # only inspect pixels at a certain distance
            # above the diagonal
            if o_j - o_i < p + min_locus_dist:
                continue

            # only inspect mappable pixels
            if mask[i, j]:
                continue

            # only inspect pixels if they have more than
            # a minimum number of reads
            w_corr = w
            ll_sum = 0
            while ll_sum < min_ll_reads and w_corr <= max_w:
                ll_sum = RaoPeakCaller.ll_sum(m_uncorrected, i, j, w=w_corr, p=p)
                if np.ma.is_masked(ll_sum):
                    ll_sum = 0
                w_corr += 1

            if w_corr > max_w:
                continue

            # calculate mappability and enrichment values
            ll_mappable = 1 - RaoPeakCaller.e_ll_sum(mask, i, j, w, p) / RaoPeakCaller.e_ll_sum(m_ones, i, j, w, p)
            if ll_mappable < min_mappable:
                continue
            e_ll = RaoPeakCaller.e_ll(m_original, i, j, m_expected, w=w_corr, p=p)

            v_mappable = 1 - RaoPeakCaller.e_v_sum(mask, i, j, w, p) / RaoPeakCaller.e_v_sum(m_ones, i, j, w, p)
            if v_mappable < min_mappable:
                continue
            e_v = RaoPeakCaller.e_v(m_original, i, j, m_expected, w=w_corr, p=p)

            h_mappable = 1 - RaoPeakCaller.e_h_sum(mask, i, j, w, p) / RaoPeakCaller.e_h_sum(m_ones, i, j, w, p)
            if h_mappable < min_mappable:
                continue
            e_h = RaoPeakCaller.e_h(m_original, i, j, m_expected, w=w_corr, p=p)

            d_mappable = 1 - RaoPeakCaller.e_d_sum(mask, i, j, w, p) / RaoPeakCaller.e_d_sum(m_ones, i, j, w, p)
            if d_mappable < min_mappable:
                continue
            e_d = RaoPeakCaller.e_d(m_original, i, j, m_expected, w=w_corr, p=p)

            # find chunks
            cf = c_i[i] * c_j[j]
            o_chunk = RaoPeakCaller.find_chunk(m_uncorrected[i, j])
            e_ll_chunk = RaoPeakCaller.find_chunk(e_ll/cf)
            e_h_chunk = RaoPeakCaller.find_chunk(e_h/cf)
            e_v_chunk = RaoPeakCaller.find_chunk(e_v/cf)
            e_d_chunk = RaoPeakCaller.find_chunk(e_d/cf)

            result = [o_i + ix_offset, o_j + ix_offset, float(m_original.data[i, j]), w_corr, p,
                      int(m_uncorrected.data[i, j]),
                      int(ll_sum), e_ll, e_v, e_h, e_d,
                      o_chunk, e_ll_chunk, e_v_chunk, e_h_chunk, e_d_chunk,
                      ll_mappable, v_mappable, h_mappable, d_mappable]

            results.append(result)
    return msgpack.dumps(results)


def overlap_peaks(peaks, max_distance=6000):
    """
    Calculate overlap between different peak calls.

    Useful for comparing peak calls across different samples
    or conditions.

    :param dict peaks: Peaks to overlap. Dictionary of
                       :class:`fanc.data.network.PeakInfo`,
                       keys are dataset names.
    :param int max_distance: Maximum distance between peaks for overlap
    :return: DataFrame of overlap statistics and dictionary
             containing overlapping peaks. Keys are sets
             of dataset names.
    :rtype: (pandas.DataFrame, fanc.data.network.PeakInfo)
    """
    # Algorithm from https://github.com/theaidenlab/juicebox/
    # blob/cb5999cb1e8e430dd29d4114fb208aca4b8d35ac/src/juicebox/
    # tools/utils/juicer/hiccups/HiCCUPSUtils.java#L235

    def key_func(p):
        try:
            return p[1].weight
        except AttributeError:
            return 1

    def hypotenuse(x, y):
        return math.sqrt(x*x + y*y)

    def mean(data):
        # https://stackoverflow.com/a/31062966
        n = 0
        mean = 0.0
        for x in data:
            n += 1
            mean += (x - mean)/n
        if n < 1:
            return float("nan")
        else:
            return mean

    summarize_attrs = [
        (mean, "weight"),
        (mean, "oe"),
        (mean, "uncorrected"),
        (mean, "expected_local"),
        (sum, "p_value"),
        (sum, "q_value_sum"),
    ]

    if not all(a == b for regions in zip(peaks.values()) for a, b in pairwise(regions)):
        raise ValueError("All peak calls must have the same regions.")

    peaks1 = next(iter(peaks.values()))
    bin_size = peaks1.bin_size
    max_distance = max_distance/bin_size
    logger.info("Fetching and sorting peaks...")

    all_peaks = list(sorted(((s, p) for s, pgen in viewitems(peaks) for p in pgen),
                            key=key_func, reverse=True))

    logger.info("Done.")
    logger.info("Finding overlaps...")
    out_peaks = defaultdict(list)
    total_n = len(all_peaks)

    with RareUpdateProgressBar(max_value=total_n, silent=config.hide_progressbars,
                               prefix="Overlap") as pb:
        while len(all_peaks) > 0:
            cur_p = all_peaks.pop(0)
            cur_p_list = [cur_p]
            cur_x = cur_p[1].x
            cur_y = cur_p[1].y
            cluster_radius = max_distance
            for p in all_peaks:
                if hypotenuse(cur_x - p[1].x, cur_y - p[1].y) <= cluster_radius:
                    cur_p_list.append(p)
                    cur_x = mean(_p.x for _s, _p in cur_p_list)
                    cur_y = mean(_p.y for _s, _p in cur_p_list)
                    r = max(hypotenuse(cur_x - _p.x, cur_y - _p.y) for _s, _p in cur_p_list)
                    cluster_radius = max_distance + r

            summed_attrs = {}
            for sum_func, attr in summarize_attrs:
                summed_attrs[attr] = sum_func(getattr(p, attr) for s, p in cur_p_list)

            cons_p = Peak(
                x=cur_x,
                y=cur_y,
                radius=r if len(cur_p_list) > 1 else 0.,
                source=math.floor(min(cur_x, cur_y)),
                sink=math.floor(max(cur_x, cur_y)),
                **summed_attrs
            )
            out_peaks[frozenset(s for s, p in cur_p_list)].append(cons_p)
            for p in cur_p_list[1:]:
                all_peaks.remove(p)
            pb.update(total_n - len(all_peaks))
    logger.info("Done.")
    logger.info("Gathering overlapped peaks.")
    out_dict = {}
    out_stats = []
    for sample_set, p_list in viewitems(out_peaks):
        pi = PeakInfo()
        pi.add_regions(peaks1.regions(), preserve_attributes=False)
        pi.add_edges(p_list)
        out_dict[sample_set] = pi
        stat = OrderedDict((s, s in sample_set) for s in peaks.keys())
        stat["n"] = len(p_list)
        out_stats.append(stat)
    logger.info("Done.")

    return pd.DataFrame(out_stats), out_dict



import numpy as np
from scipy.stats.mstats import gmean


def remove_sparse_rows(m, cutoff=None):
    s = np.sum(m, 0)
    
    if cutoff is None:
        cutoff = min(s)
    
    idxs = np.where(s <= cutoff)[0]
    m_removed = np.delete(m, idxs, 0)
    m_removed = np.delete(m_removed, idxs, 1)
    
    return m_removed, idxs
    
    
def restore_sparse_rows(m, idx_sets, rows=None):
    abs_idx = []
    for idxs in reversed(idx_sets):
        for i in sorted(idxs):
            shift = 0
            for j in sorted(abs_idx):
                if j + shift < i:
                    shift += 1
            abs_idx.append(i - shift)
    abs_idx.sort()
    a = np.insert(m, abs_idx, 0, axis=0)
    if len(m.shape) > 1:
        a = np.insert(a, abs_idx, 0, axis=1)
    return a


def is_symmetric(m, tol=1e-10):
    for i in range(0, m.shape[0]):
        for j in range(i, m.shape[1]):
            if abs(m[i, j]-m[j, i]) > tol:
                return False
    return True


def apply_sliding_func(a, window, func=np.ma.mean):
    """
    Apply function on a sliding window over an array, ignoring Numpy NaN values.

    :param a: Numpy array on which function is applied
    :param window: The sliding window is i - window:i + window + 1
                   so total window is twice this parameter.
    :param func: Function to apply
    """
    out = np.empty(a.shape)
    for i in range(len(a)):
        window_start = max(0, i - window)
        window_end = min(len(a), i + window + 1)
        cur_window = a[window_start:window_end]
        out[i] = func(cur_window[~np.isnan(cur_window)])
    return out


def delta_window(x, window, ignore_mask=False, mask_thresh=.5):
    try:
        x.mask = np.logical_or(x.mask, ~np.isfinite(x))
    except AttributeError:
        x = np.ma.masked_invalid(x)
    n = len(x)
    delta = np.empty(n)
    for i in range(n):
        if i < window or n - i <= window - 1:
            delta[i] = np.nan
            continue
        down_slice = slice(i + 1, i + window + 1)
        up_slice = slice(i - window, i)
        if not ignore_mask and (np.sum(x.mask[down_slice]) > window*mask_thresh or
                np.sum(x.mask[up_slice]) > window*mask_thresh):
            delta[i] = np.nan
        delta[i] = np.ma.mean(x[down_slice] - x[i]) - np.ma.mean(x[up_slice] - x[i])
    return delta


def kth_diag_indices(n, k):
    """
    Return indices of bins k steps away from the diagonal.
    (from http://stackoverflow.com/questions/10925671/numpy-k-th-diagonal-indices)
    """

    rows, cols = np.diag_indices(n)
    if k < 0:
        return rows[-k:], cols[:k]
    elif k > 0:
        return rows[:-k], cols[k:]
    else:
        return rows, cols


def trim_stats(a, proportiontocut=0.0, axis=0, stat=np.nanmean):
    if proportiontocut >= 0.5:
        raise ValueError("Cannot cut more than 50% of values off distribution tails!")

    s = np.sort(a, axis=axis)
    ix = int(s.shape[axis] * proportiontocut)

    sl = [slice(None)] * s.ndim
    sl[axis] = slice(ix, len(s)-ix, 1)
    s_sub = s[tuple(sl)]

    return stat(s_sub)


def nangmean(a, axis=0, dtype=None):
    return gmean(a[np.isfinite(a)], axis=axis, dtype=dtype)



import logging

import os
import h5py
import numpy as np
import pandas
import cooler
from genomic_regions import GenomicRegion
from ..tools.sambam import natural_cmp
from functools import cmp_to_key
import tempfile
import warnings
from ..tools.files import tmp_file_name
import shutil


from ..matrix import RegionMatrixContainer, Edge

logger = logging.getLogger(__name__)


def is_cooler(file_name):
    try:
        if "@" in file_name:
            hic_file, at_resolution = file_name.split("@")
            file_name = hic_file + '::resolutions/{}'.format(at_resolution)
        cooler.Cooler(file_name)
        return True
    except KeyError:
        return False


def to_cooler(hic, path, balance=True, multires=True,
              resolutions=None, n_zooms=10, threads=1,
              chunksize=100000, max_resolution=5000000,
              natural_order=True, chromosomes=None,
              **kwargs):
    """
    Export Hi-C data as Cooler file.

    Only contacts that have not been
    filtered are exported. https://github.com/mirnylab/cooler/

    Single resolution files:
    If input Hi-C matrix is uncorrected, the uncorrected matrix is stored.
    If it is corrected, the uncorrected matrix is stored along with bias vector.
    Cooler always calculates corrected matrix on-the-fly from the uncorrected
    matrix and the bias vector.

    Multi-resolution files (default):


    :param hic: Hi-C file in any compatible (RegionMatrixContainer) format
    :param path: Output path for cooler file
    :param balance: Include bias vector in cooler output (single res) or perform
                    iterative correction (multi res)
    :param multires: Generate a multi-resolution cooler file
    :param resolutions: Resolutions in bp (int) for multi-resolution cooler output
    :param chunksize: Number of pixels processed at a time in cooler
    :param kwargs: Additional arguments passed to cooler.iterative_correction
    """
    base_resolution = hic.bin_size

    tmp_files = []
    try:
        if multires:
            if resolutions is None:
                resolutions = [base_resolution * 2 ** i for i in range(n_zooms)
                               if base_resolution * 2 ** i < max_resolution]
            else:
                for r in resolutions:
                    if r % base_resolution != 0:
                        raise ValueError("Resolution {} must be a multiple of "
                                         "base resolution {}!".format(r, base_resolution))

            single_path = tempfile.NamedTemporaryFile(delete=False, suffix='.cool').name
            tmp_files.append(single_path)
            multi_path = path
        else:
            single_path = path
            multi_path = None

        natural_key = cmp_to_key(natural_cmp)
        if chromosomes is None:
            chromosomes = hic.chromosomes()
            if natural_order:
                chromosomes = sorted(chromosomes, key=lambda x: natural_key(x.encode('utf-8')))

        logger.info("Loading genomic regions")
        ix_converter = dict()
        regions = []
        region_order = []
        new_region_index = 0
        for chromosome in chromosomes:
            for region in hic.regions(chromosome, lazy=True):
                regions.append((region.chromosome,
                                region.start - 1,
                                region.end))
                ix_converter[region.ix] = new_region_index
                region_order.append(region.ix)
                new_region_index += 1
        region_df = pandas.DataFrame(regions, columns=['chrom', 'start', 'end'])

        def pixel_iter():
            for chri in range(len(chromosomes)):
                chromosome1 = chromosomes[chri]
                for chrj in range(chri, len(chromosomes)):
                    chromosome2 = chromosomes[chrj]

                    logger.info("{} - {}".format(chromosome1, chromosome2))

                    def chromosome_pixel_iter():
                        for edge in hic.edges((chromosome1, chromosome2), norm=False, lazy=True):
                            source, sink = ix_converter[edge.source], ix_converter[edge.sink]
                            if sink < source:
                                source, sink = sink, source
                            yield source, sink, edge.weight

                    pixels = np.fromiter(chromosome_pixel_iter(),
                                         dtype=[("bin1_id", np.int_),
                                                ("bin2_id", np.int_),
                                                ("count", np.float_)])
                    pixels = np.sort(pixels, order=("bin1_id", "bin2_id"))
                    if len(pixels) > 0:
                        yield pandas.DataFrame(pixels)

        logger.info("Writing cooler")
        cooler.create_cooler(cool_uri=single_path, bins=region_df, pixels=pixel_iter(), ordered=False)

        cool_path, group_path = cooler.util.parse_cooler_uri(single_path)

        if not multires:
            if balance:
                logger.info("Writing bias vector from FAN-C matrix")
                bias = hic.bias_vector()[np.array(region_order)]

                # Copied this section from
                # https://github.com/mirnylab/cooler/blob/356a89f6a62e2565f42ff13ec103352f20d251be/cooler/cli/balance.py#L195
                with h5py.File(cool_path, 'r+') as h5:
                    grp = h5[group_path]
                    # add the bias column to the file
                    h5opts = dict(compression='gzip', compression_opts=6)
                    grp['bins'].create_dataset("weight", data=bias, **h5opts)
            return CoolerHic(single_path)
        else:
            cooler.zoomify_cooler(single_path, multi_path, resolutions, chunksize, nproc=threads)
            if balance:
                logger.info("Balancing zoom resolutions...")
                for resolution in resolutions:
                    uri = multi_path + "::resolutions/" + str(resolution)
                    cool_path, group_path = cooler.util.parse_cooler_uri(uri)
                    cool = cooler.Cooler(uri)
                    bias, stats = cooler.balance_cooler(cool, chunksize=chunksize, **kwargs)
                    with h5py.File(cool_path, 'r+') as h5:
                        grp = h5[group_path]
                        # add the bias column to the file
                        h5opts = dict(compression='gzip', compression_opts=6)
                        grp['bins'].create_dataset("weight", data=bias, **h5opts)
                        grp['bins']['weight'].attrs.update(stats)
            return CoolerHic(multi_path + '::resolutions/{}'.format(base_resolution))
    finally:
        for tmp_file in tmp_files:
            os.remove(tmp_file)


class LazyCoolerEdge(object):
    def __init__(self, series, c):
        self._series = series
        self._c = c
        self._weight_field = 'weight'
        self.bias = 1.
        self.expected = None

    def __getattr__(self, item):
        return self._row[item]

    @property
    def weight(self):
        if self.expected is None:
            return float(self._series['count']) * self.bias
        else:
            return (float(self._series['count']) * self.bias) / self.expected

    def __getitem__(self, item):
        try:
            return getattr(self, item)
        except AttributeError:
            raise KeyError("No such key: {}".format(item))

    @property
    def source(self):
        return int(self._series.bin1_id)

    @property
    def sink(self):
        return int(self._series.bin2_id)

    @property
    def source_node(self):
        return self._c.regions[self.source]

    @property
    def sink_node(self):
        return self._c.regions[self.sink]

    @property
    def field_names(self):
        return ['weight']


class LazyCoolerRegion(GenomicRegion):
    def __init__(self, series, ix=None):
        self._series = series
        self.ix = ix

    def __getattr__(self, item):
        return getattr(self._series, item)

    @property
    def chromosome(self):
        try:
            return self._series.chrom
        except Exception as e:
            print(e)
            raise

    @property
    def start(self):
        return self._series.start + 1

    @property
    def bias(self):
        return self._series.weight

    @property
    def strand(self):
        try:
            return self._series.strand
        except AttributeError:
            return 1


class CoolerHic(RegionMatrixContainer, cooler.Cooler):
    def __init__(self, *args, **kwargs):
        mode = kwargs.pop("mode", 'r')
        tmpdir = kwargs.pop('tmpdir', None)

        if mode != 'r':
            warnings.warn("Mode {} not compatible with CoolerHic. "
                          "File will be opened in read-only mode and "
                          "changes will not be saved to file!")
        largs = list(args)
        at_resolution = None

        hic_file = largs[0]
        if "@" in largs[0]:
            hic_file, at_resolution = largs[0].split("@")

        if tmpdir is None or (isinstance(tmpdir, bool) and not tmpdir):
            self.tmp_file_name = None
        else:
            logger.info("Working in temporary directory...")
            if isinstance(tmpdir, bool):
                tmpdir = tempfile.gettempdir()
            else:
                tmpdir = os.path.expanduser(tmpdir)
            self.tmp_file_name = tmp_file_name(tmpdir, prefix='tmp_fanc', extension='.mcool')
            logger.info("Temporary file: {}".format(self.tmp_file_name))
            shutil.copyfile(hic_file, self.tmp_file_name)
            hic_file = self.tmp_file_name

        if at_resolution is None:
            largs[0] = hic_file
        else:
            largs[0] = hic_file + '::resolutions/{}'.format(at_resolution)

        cooler.Cooler.__init__(self, *largs, **kwargs)
        RegionMatrixContainer.__init__(self)
        self._mappability = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return exc_type is None

    def close(self, remove_tmp=True):
        """
        Close this Juicer file and run exit operations.

        If file was opened with tmpdir in read-only mode:
        close file and delete temporary copy.

        :param remove_tmp: If False, does not delete temporary copy of file.
        """
        if self.tmp_file_name is not None and remove_tmp:
            os.remove(self.tmp_file_name)

    def _series_to_region(self, series, ix=None, lazy_region=None):
        if lazy_region is not None:
            lazy_region._series = series
            lazy_region.ix = ix
            return lazy_region

        index = set(series.index)
        index.remove('chrom')
        try:
            index.remove('ix')
        except KeyError:
            pass

        kwargs = {name: series[name] for name in index}
        kwargs['chromosome'] = series.chrom
        try:
            kwargs['bias'] = series.weight
        except AttributeError:
            kwargs['bias'] = 1.
        kwargs['start'] = series.start + 1
        if ix is not None:
            kwargs['ix'] = ix
        return GenomicRegion(**kwargs)

    def _region_iter(self, lazy=False, *args, **kwargs):
        lazy_region = LazyCoolerRegion(None) if lazy else None

        for df in self.bins():
            yield self._series_to_region(df.iloc[0], ix=df.index[0], lazy_region=lazy_region)

    def _region_subset(self, region, lazy=False, *args, **kwargs):
        lazy_region = LazyCoolerRegion(None) if lazy else None

        query = "{}".format(region.chromosome)
        if region.start is not None and region.end is not None:
            query += ':{}-{}'.format(region.start - 1, region.end)

        df = self.bins().fetch(query)
        for index, row in df.iterrows():
            yield self._series_to_region(row, ix=index, lazy_region=lazy_region)

    def _get_regions(self, item, *args, **kwargs):
        regions = []
        df = self.bins()[item]
        for index, row in df.iterrows():
            regions.append(self._series_to_region(row, ix=index))

        if isinstance(item, int):
            return regions[0]
        return regions

    def _region_len(self):
        return len(self.bins())

    def chromosomes(self):
        cs = []
        for index, row in self.chroms()[:].iterrows():
            cs.append(row['name'])
        return cs

    @property
    def chromosome_lengths(self):
        cl = {}
        for index, row in self.chroms()[:].iterrows():
            cl[row['name']] = row['length']
        return cl

    def _chromosome_bins(self, *args, **kwargs):
        return RegionMatrixContainer._chromosome_bins(self, lazy=True)

    def _series_to_edge(self, series, lazy_edge=None):
        if lazy_edge is None:
            index = set(series.index)
            index.remove('bin1_id')
            index.remove('bin2_id')
            index.remove('count')

            kwargs = {name: series[name] for name in index}
            kwargs['source'] = int(series['bin1_id'])
            kwargs['sink'] = int(series['bin2_id'])
            kwargs['weight'] = float(series['count'])
            kwargs['source_node'] = self.regions[kwargs['source']]
            kwargs['sink_node'] = self.regions[kwargs['sink']]

            return Edge(**kwargs)
        else:
            lazy_edge._series = series
            return lazy_edge

    def _edges_iter(self, lazy=False, *args, **kwargs):
        lazy_edge = LazyCoolerEdge(None, self) if lazy else None
        selector = cooler.Cooler.matrix(self, as_pixels=True, balance=False)
        for df in selector:
            for index, row in df.iterrows():
                yield self._series_to_edge(row, lazy_edge=lazy_edge)

    def _edges_subset(self, key=None, row_regions=None, col_regions=None,
                      lazy=False, *args, **kwargs):
        lazy_edge = LazyCoolerEdge(None, self) if lazy else None
        row_start, row_end = self._min_max_region_ix(row_regions)
        col_start, col_end = self._min_max_region_ix(col_regions)

        df = cooler.Cooler.matrix(self, as_pixels=True, balance=False)[row_start:row_end+1, col_start:col_end+1]

        for index, row in df.iterrows():
            yield self._series_to_edge(row, lazy_edge=lazy_edge, *args, **kwargs)

    def _edges_getitem(self, item, *args, **kwargs):
        edges = []
        df = self.pixels()[item]
        for index, row in df.iterrows():
            edges.append(self._series_to_edge(row))

        if isinstance(item, int):
            return edges[0]
        return edges

    def _edges_length(self):
        return len(self.pixels())

    def mappable(self, region=None):
        """
        Get the mappability vector of this matrix.
        """
        if self._mappability is not None:
            return self._mappability

        mappable = [False] * len(list(self.regions(region)))
        for edge in self.edges(region, lazy=True):
            mappable[edge.source] = True
            mappable[edge.sink] = True
        self._mappability = mappable
        return mappable

    def bias_vector(self):
        return np.array([r.bias for r in self.regions(lazy=True)])



"""
This module performs iterative mapping of reads in FASTQ files to a reference genome.

The main function is :func:`~iterative_mapping`, which requires an input FASTQ file,
an output SAM file, and a suitable mapper. Mapper options are :class:`~Bowtie2Mapper`
and :class:`~BwaMapper` for the time being, but by subclassing :class:`~Mapper` a user
can easily write their own mapper implementations that can fully leverage the iterative
mapping capabilities of FAN-C. Take a look at the code of :class:`~Bowtie2Mapper` for
an example.

Example usage:

.. code::

    import fanc
    mapper = fanc.BwaMapper("bwa-index/hg19_chr18_19.fa", min_quality=3)
    fanc.iterative_mapping("SRR4271982_chr18_19_1.fastq.gzip", "SRR4271982_chr18_19_1.bam",
                           mapper, threads=4, restriction_enzyme="HindIII")

"""

import os
import io
import subprocess
import multiprocessing as mp
from queue import Empty, Full
import traceback
import gzip
import re
import threading
import pysam
import tempfile
import shutil
import uuid
from future.utils import string_types
from .config import config
from .tools.files import which
from .tools.general import ligation_site_pattern, split_at_ligation_junction, WorkerMonitor
import logging
logger = logging.getLogger(__name__)

__all__ = ['Bowtie2Mapper', 'BwaMapper',
           'SimpleBowtie2Mapper', 'SimpleBwaMapper',
           'iterative_mapping']

_read_name_re = re.compile("^@(.+?)\s(.*)$")
_read_name_nospace_re = re.compile("^@(.+)$")


def read_name(line):
    """
    Extract read name from FASTQ line.

    :param line: str of the FASTQ line, starting with "@"
    :return: tuple of read name and other info. info can be None
    """
    matches = _read_name_re.match(line)
    if matches is None:
        matches = _read_name_nospace_re.match(line)
        name = matches.group(1)
        info = None
    else:
        name = matches.group(1)
        info = matches.group(2)
    return name, info


class Monitor(WorkerMonitor):
    """
    Monitor class keeping an eye on mapping workers.

    Thread-safe by using locks on each attribute.

    .. attribute:: value

        Counter used to track worker progress.
    """
    def __init__(self, value=0, manager=None):
        WorkerMonitor.__init__(self, value=value, manager=manager)

        self.resubmitting_lock = manager.Lock()
        self.submitting_lock = manager.Lock()

        with self.resubmitting_lock:
            self.resubmitting = False

        with self.submitting_lock:
            self.submitting = False

    def set_resubmitting(self, value):
        """
        Set the resubmission status of a mapping worker.

        :param value: True if worker status changed to resubmitting,
                      False otherwise
        """
        with self.resubmitting_lock:
            self.resubmitting = value

    def is_resubmitting(self):
        """
        Query resubission status of a worker.

        :return: True if worker is in the process of resubmitting,
                 False otherwise
        """
        with self.resubmitting_lock:
            return self.resubmitting

    def set_submitting(self, value):
        """
        Set the submission status of a mapping worker.

        :param value: True if worker status changed to submitting,
                      False otherwise
        """
        with self.submitting_lock:
            self.submitting = value

    def is_submitting(self):
        """
        Query subission status of a worker.

        :return: True if worker is in the process of submitting,
                 False otherwise
        """
        with self.submitting_lock:
            return self.submitting


class Mapper(object):
    """
    Mapper base class for mapping sequencing reads to a reference genome.

    This base class handles the resubmission checks for iterative mapping,
    spawns mapping processes, and compiles a FASTQ that can be used as input
    for external mapping software.

    To actually do any work, this class must be inherited, and the _map and
    _resubmit methods must be overridden.

    :func:`~Mapper._map` must handle the actual mapping given an input file

    :func:`~Mapper._resubmit` determines whether a read should be resubmitted
    (after expanding it)

    See :class:`~Bowtie2Mapper` or :class:`~BwaMapper` for examples.

    .. attribute:: resubmit_unmappable

        Set to True (default) if unmappable reads should be resubmitted

    .. attribute:: attempt_resubmit

        Try to resubmit a read if it failed the :func:`~Mapper._resubmit`
        test.
    """

    def __init__(self, resubmit_unmappable=True, attempt_resubmit=True):
        self.resubmit_unmappable = resubmit_unmappable
        self.attempt_resubmit = attempt_resubmit

    def _map(self, input_file, output_file, *args, **kwargs):
        """
        Map reads in the given FASTQ file to a reference genome.

        Must be overridden by classes implementing :class:`~Mapper`.
        Arguments are provided at runtime by the iterative mapping
        process.

        :param input_file: Path to FASTQ file
        :param output_file: Path to SAM file
        :param args: Optional positional arguments
        :param kwargs: Optional keyword arguments
        :return: Must return 0 if mapping successful,
                 any other integer otherwise
        """
        raise NotImplementedError("Must implement _map method!")

    def map(self, input_file, output_folder=None):
        """
        Map reads in the given FASTQ file using :func:`~Mapper._map` implementation.

        Will internally map the FASTQ reads to a SAM file in a
        temporary folder (use output_folder to choose a specific folder),
        split the SAM output into (i) valid alignments according to
        :func:`~Mapper._resubmit` and (ii) invalid alignments that get
        caught by the resubmission filter. A FASTQ file will be constructed
        from the invalid alignments, extending the reads by a given step size,
        which can then be used to iteratively repeat the mapping process until
        a valid alignment is found or the full length of the read has been
        restored.

        :param input_file: Path to FASTQ file
        :param output_folder: (optional) path to temporary folder for SAM output
        :return: tuple, path to valid SAM alignments, path to resubmission FASTQ
        """
        if output_folder is None:
            output_folder = tempfile.mkdtemp()

        logger.debug('Output folder for SAM process: {}'.format(output_folder))

        with tempfile.NamedTemporaryFile(prefix='output', suffix='.sam', dir=output_folder,
                                         delete=False) as tmp:
            sam_output_file = tmp.name

            ret = self._map(input_file, sam_output_file)

            if isinstance(ret, tuple) and len(ret) == 2:
                ret, error_message = ret
            else:
                error_message = ''

        if ret != 0:
            raise RuntimeError('Mapping had non-zero exit status {}. {}'.format(ret, error_message))

        logger.debug('Done mapping')

        if not self.resubmit_unmappable and self.attempt_resubmit:
            sam_valid_file, resubmission_file = sam_output_file, None
        else:
            sam_valid_file, resubmit = self._valid_and_resubmissions(sam_output_file, output_folder)
            os.remove(sam_output_file)
            resubmission_file = self._resubmission_fastq(input_file, resubmit, output_folder)

        logger.debug('Mapper done.')
        return sam_valid_file, resubmission_file

    def _resubmit(self, sam_fields):
        """
        Determine if an alignment should be resubmitted.

        :param sam_fields: The individual fields in a SAM line (split by tab)
        :return: True if read should be extended and aligned to the reference again,
                 False if the read passes the validity criteria.
        """
        raise NotImplementedError("Mapper must implement '_resubmit'")

    def resubmit(self, sam_fields):
        """
        Determine if an alignment should be resubmitted.

        Filters unmappable reads by default. Additional criteria can be
        implemented using the :func:`~Mapper._resubmit` method.

        :param sam_fields: The individual fields in a SAM line (split by tab)
        :return: True if read should be extended and aligned to the reference again,
                 False if the read passes the validity criteria.
        """
        if self.resubmit_unmappable and int(sam_fields[1]) & 4:
            return True
        return self._resubmit(sam_fields)

    def _valid_and_resubmissions(self, input_sam_file, output_folder):
        """
        Determine valid alignments and resubmissions from SAM file.

        :param input_sam_file: Path to SAM file
        :param output_folder: Path to temporary output folder
        :return: tuple with path to valid SAM file, dictionary with
                 read names that should be resubmitted
        """
        logger.debug('Getting mapped reads and resubmissions')
        resubmit = dict()
        with io.open(input_sam_file) as f:
            with tempfile.NamedTemporaryFile(mode='w', prefix='valid', suffix='.sam',
                                             delete=False, dir=output_folder) as o:
                output_sam_file = o.name
                for i, line in enumerate(f):
                    line = line.rstrip()
                    if line == '':
                        continue
                    if line.startswith('@'):
                        o.write(line + '\n')
                        continue

                    sam_fields = line.split("\t")

                    if self._resubmit(sam_fields):
                        if not sam_fields[0] in resubmit:
                            resubmit[sam_fields[0]] = True
                    else:
                        resubmit[sam_fields[0]] = False
                        o.write(line + '\n')

        return output_sam_file, resubmit

    def _resubmission_fastq(self, input_fastq, resubmit, output_folder):
        """
        Create the FASTQ file for resubmissions.

        :param input_fastq: Path to original FASTQ file
        :param resubmit: Dictionary with read names that should be expanded
                         and then resubmitted
        :param output_folder: temporary folder
        :return: Path to resubmission FASTQ
        """
        logger.debug('Creating FASTQ file for resubmission')

        resubmission_counter = 0
        total_counter = 0
        with io.open(input_fastq) as f:
            with tempfile.NamedTemporaryFile(mode='w', prefix='resubmission_', suffix='.fastq',
                                             delete=False, dir=output_folder) as o:
                output_fastq = o.name
                resubmit_current = False
                for i, line in enumerate(f):
                    line = line.rstrip()
                    if line == '':
                        continue

                    if i % 4 == 0:
                        name, _ = read_name(line)
                        if name not in resubmit:
                            if self.resubmit_unmappable:
                                resubmit_current = True
                        elif resubmit[name]:
                            resubmit_current = True
                        else:
                            resubmit_current = False

                    if resubmit_current:
                        resubmission_counter += 1
                        o.write(line + '\n')
                    total_counter += 1

        if resubmission_counter == 0:
            os.remove(output_fastq)
            return None

        logger.debug("Resubmitting {}/{}".format(resubmission_counter/4, total_counter/4))
        return output_fastq

    def close(self):
        """
        Final operations after mapping completes.
        """
        pass


class Bowtie2Mapper(Mapper):
    """
    Bowtie2 Mapper for aligning reads against a reference genome.

    Implements :class:`~Mapper` by calling the command line "bowtie2"
    program.

    .. attribute:: bowtie2_index

        Path to the bowtie2 index of the reference genome of choice.

    .. attribute:: min_quality

        Minimum MAPQ of an alignment so that it won't be resubmitted in
        iterative mapping.

    .. attribute:: additional_arguments

        Arguments passed to the "bowtie2" command in addition to
        -x, -U, --no-unal, --threads, and -S.

    .. attribute:: threads

        Number of threads for this mapping process.

    """
    def __init__(self, bowtie2_index, min_quality=30, additional_arguments=(),
                 threads=1, _bowtie2_path=config.bowtie2_path, **kwargs):
        Mapper.__init__(self, **kwargs)
        self.index = os.path.expanduser(bowtie2_index)
        if self.index.endswith('.'):
            self.index = self.index[:-1]
        self.args = [a for a in additional_arguments]
        self._path = _bowtie2_path
        if self._path is None:
            self._path = "bowtie2"
        if which(self._path) is None:
            raise ValueError("Cannot find {}".format(self._path))
        self.min_quality = min_quality
        self.threads = threads
        self.attempt_resubmit = (self.min_quality is not None and self.min_quality > 0)

    def _map(self, input_file, output_file, *args, **kwargs):
        bowtie2_command = [self._path, '-x', self.index, '-U', input_file, '--no-unal',
                           '--threads', str(self.threads), '-S', output_file] + self.args
        logger.debug('Bowtie2 command: {}'.format(bowtie2_command))

        proc = subprocess.Popen(bowtie2_command, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, universal_newlines=True)
        proc.wait()

        if proc.returncode != 0:
            error_message = proc.stderr.read()
        else:
            error_message = ''

        return proc.returncode, error_message

    def _resubmit(self, sam_fields):
        if int(sam_fields[4]) < self.min_quality:
            return True
        return False


class SimpleBowtie2Mapper(Bowtie2Mapper):
    """
    Bowtie2 Mapper for aligning reads against a reference genome without resubmission.

    Implements :class:`~Mapper` by calling the command line "bowtie2"
    program. Does not resubmit reads under any circumstance.

    .. attribute:: bowtie2_index

        Path to the bowtie2 index of the reference genome of choice.

    .. attribute:: additional_arguments

        Arguments passed to the "bowtie2" command in addition to
        -x, -U, --no-unal, --threads, and -S.

    .. attribute:: threads

        Number of threads for this mapping process.
    """
    def __init__(self, bowtie2_index, additional_arguments=(),
                 threads=1, _bowtie2_path='bowtie2'):
        Bowtie2Mapper.__init__(self, bowtie2_index, min_quality=0,
                               additional_arguments=additional_arguments,
                               threads=threads,
                               _bowtie2_path=_bowtie2_path,
                               resubmit_unmappable=False)

    def _resubmit(self, sam_fields):
        return False


class BwaMapper(Mapper):
    """
    BWA Mapper for aligning reads against a reference genome.

    Implements :class:`~Mapper` by calling the command line "bwa"
    program.

    .. attribute:: bwa_index

        Path to the BWA index of the reference genome of choice.

    .. attribute:: min_quality

        Minimum MAPQ of an alignment so that it won't be resubmitted in
        iterative mapping.

    .. attribute:: additional_arguments

        Arguments passed to the "bowtie2" command in addition to
        -t and -o.

    .. attribute:: threads

        Number of threads for this mapping process.

    .. attribute:: algorithm

        BWA algorithm to use for mapping. Uses "mem" by default.
        See http://bio-bwa.sourceforge.net/bwa.shtml for other options.

    .. attribute: memory_map

        Set to true if you want to commit the entire BWA index to memory.
        This sometimes leads to clashes between different BWA threads.

    """
    def __init__(self, bwa_index, min_quality=0, additional_arguments=(),
                 threads=1, algorithm='mem', memory_map=False, _bwa_path=config.bwa_path):
        Mapper.__init__(self)
        self.index = os.path.expanduser(bwa_index)
        if self.index.endswith('.'):
            self.index = self.index[:-1]
        self.args = [a for a in additional_arguments]
        self._path = _bwa_path
        if self._path is None:
            self._path = 'bwa'
        if which(self._path) is None:
            raise ValueError("Cannot find {}".format(self._path))
        self.algorithm = algorithm
        self.min_quality = min_quality
        self.threads = threads
        self.attempt_resubmit = (self.min_quality is not None and self.min_quality > 0)
        self.memory_map = memory_map

    def _map(self, input_file, output_file, *args, **kwargs):
        if self.memory_map:
            with open(os.devnull, 'w') as f_null:
                ret = subprocess.call(['bwa', 'shm', self.index], stderr=subprocess.STDOUT, stdout=f_null)
                if ret == 0:
                    logger.debug("Memory mapped BWA index")

        bwa_command = [self._path, self.algorithm, '-t', str(self.threads), '-o', output_file] + \
                      self.args + \
                      [self.index, input_file]
        logger.debug('BWA command: {}'.format(bwa_command))

        proc = subprocess.Popen(bwa_command, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, universal_newlines=True)
        proc.wait()

        if proc.returncode != 0:
            error_message = proc.stderr.read()
        else:
            error_message = ''

        return proc.returncode, error_message

    def _resubmit(self, sam_fields):
        if int(sam_fields[4]) < self.min_quality:
            return True
        return False

    def close(self):
        logger.debug("Deleting shared memory index")
        try:
            with open(os.devnull, 'w') as f_null:
                ret = subprocess.call(['pgrep', '-x', 'bwa'], stderr=subprocess.STDOUT, stdout=f_null)
                if ret == 1:
                    ret2 = subprocess.call(['bwa', 'shm', '-d'], stderr=subprocess.STDOUT, stdout=f_null)
                    if ret2 == 0:
                        logger.debug("Deleted memory mapped BWA index")

        except FileNotFoundError:
            pass


class SimpleBwaMapper(BwaMapper):
    """
    BWA Mapper for aligning reads against a reference genome without resubmission.

    Implements :class:`~Mapper` by calling the command line "bwa"
    program. Does not resubmit reads under any circumstance, i.e.
    does not perform iterative mapping.

    .. attribute:: bwa_index

        Path to the BWA index of the reference genome of choice.

    .. attribute:: min_quality

        Minimum MAPQ of an alignment so that it won't be resubmitted in
        iterative mapping.

    .. attribute:: additional_arguments

        Arguments passed to the "bowtie2" command in addition to
        -t and -o.

    .. attribute:: threads

        Number of threads for this mapping process.

    .. attribute:: algorithm

        BWA algorithm to use for mapping. Uses "mem" by default.
        See http://bio-bwa.sourceforge.net/bwa.shtml for other options.

    .. attribute: memory_map

        Set to true if you want to commit the entire BWA index to memory.
        This sometimes leads to clashes between different BWA threads.

    """
    def __init__(self, bwa_index, additional_arguments=(),
                 threads=1, memory_map=False, _bwa_path='bwa'):
        BwaMapper.__init__(self, bwa_index, min_quality=0,
                           additional_arguments=additional_arguments,
                           threads=threads, memory_map=memory_map,
                           _bwa_path=_bwa_path)
        self.resubmit_unmappable = False
        self.attempt_resubmit = False

    def _resubmit(self, sam_fields):
        return False


def _trim_read(input_seq, step_size=5, min_size=25, front=False):
    """
    Trim an input sequence by step_size.

    :param input_seq: input DNA sequence
    :param step_size: truncation step size
    :param min_size: minimum size of input DNA before error is raised
    :param front: If True, trim from front instead of back.
    :return: tuple: read name, trimmed sequence, '+', and trimmed quality
                    string
    """
    name, seq, plus, qual = input_seq
    if len(seq) <= min_size:
        raise ValueError("Already reached minimum size, cannot truncate read further")

    if len(seq) - step_size < min_size:
        final_length = min_size
    else:
        final_length = len(seq) - step_size

    if front:
        sl = len(seq)
        seq = seq[sl - final_length:]
        qual = qual[sl - final_length:]
    else:
        seq = seq[:final_length]
        qual = qual[:final_length]
    return name, seq, plus, qual


def _iterative_mapping_worker(mapper, input_queue, output_folder, output_queue,
                              resubmission_queue, monitor, exception_queue):
    """
    Worker performing the iterative mapping of FASTQ files.

    :param mapper: :class:`~Mapper` implementation
    :param input_queue: queue with input FASTQ files
    :param output_folder: Output folder for intermediate results
    :param output_queue: queue for valid output alignments
    :param resubmission_queue: queue for resubmissions (iterative mapping)
    :param monitor: worker monitor
    :param exception_queue: queue for exceptions
    """
    try:
        # generate worker's uuid
        worker_uuid = uuid.uuid4()

        while True:
            # set worker state to idle
            monitor.set_worker_idle(worker_uuid)
            logger.debug('Worker {} idle'.format(worker_uuid))

            # wait for input
            input_file = input_queue.get(True)
            monitor.set_worker_busy(worker_uuid)
            logger.debug('Mapper {} busy, got input file'.format(worker_uuid))

            # mapping file
            sam_file, unmapped_file = mapper.map(input_file, output_folder=output_folder)
            logger.debug('{} done mapping'.format(worker_uuid))

            # clean up
            os.remove(input_file)
            logger.debug('{} waiting to put SAM in output queue'.format(worker_uuid))
            output_queue.put(sam_file)

            # send resubmissions back to writing thread
            logger.debug('{} waiting to put FASTQ in resubmission queue'.format(worker_uuid))
            resubmission_queue.put(unmapped_file)
            logger.debug('{} finished mapping round.'.format(worker_uuid))
    except Exception:
        import sys
        exception_queue.put("".join(traceback.format_exception(*sys.exc_info())))


def _fastq_to_queue(fastq_file, output_folder, batch_size, input_queue, monitor,
                    exception_queue=None, worker_pool=None, restriction_enzyme=None):
    """
    Submit FASTQ batches to input queue.

    Also does ligation site splitting, if requested.

    :param fastq_file: Path to input FASTQ file
    :param output_folder: Path to output folder (tmp)
    :param batch_size: Number of reads submitted to a monitor
    :param input_queue: queue for batch FASTQ files
    :param monitor: worker monitor
    :param exception_queue: queue for exceptions
    :param worker_pool: Pool of mapping workers "_iterative_mapping_worker"
    :param restriction_enzyme: Name of restriction enzyme of restriction pattern
                               (e.g. A^AGCT_T)
    """
    monitor.set_submitting(True)

    if restriction_enzyme is not None:
        if isinstance(restriction_enzyme, string_types):
            restriction_enzyme = [restriction_enzyme]

        ligation_patterns = [ligation_site_pattern(r) for r in restriction_enzyme]
    else:
        ligation_patterns = None

    tmp_output_file = None
    line_counter = 0
    submission_counter = 0
    try:
        if fastq_file.endswith('.gz') or fastq_file.endswith('.gzip'):
            open_file = lambda x: io.BufferedReader(gzip.open(x, 'r'),
                                                    buffer_size=4*io.DEFAULT_BUFFER_SIZE)
        else:
            open_file = io.open

        read_counter = 0
        with open_file(fastq_file) as f:
            tmp_output_file = tempfile.NamedTemporaryFile(suffix='.fastq', delete=False,
                                                          dir=output_folder, mode='w+b')

            current_fastq = []
            for i, line in enumerate(f):
                line = line.encode('utf-8') if not isinstance(line, bytes) else line
                current_fastq.append(line)
                line_counter += 1

                # line = line.decode() if isinstance(line, bytes) else line
                if line_counter % 4 == 0:
                    if not ligation_patterns:
                        for fastq_line in current_fastq:
                            tmp_output_file.write(fastq_line)
                        read_counter += 1
                    else:
                        name, info = read_name(current_fastq[0].rstrip().decode())
                        seq = current_fastq[1].rstrip()
                        qual = current_fastq[3].rstrip()
                        seqs = split_at_ligation_junction(seq.decode(), ligation_patterns)

                        qualities = []
                        current_pos = 0
                        for s in seqs:
                            qualities.append(qual[current_pos:current_pos+len(s)])
                            current_pos += len(s)

                        for j in range(len(seqs)):
                            if len(seqs) > 1:
                                new_name = '@' + name + '__{}'.format(j)
                            else:
                                new_name = '@' + name

                            if info is not None:
                                new_name += ' ' + info
                            tmp_output_file.write(new_name.encode('utf-8') + b'\n')
                            tmp_output_file.write(seqs[j].encode('utf-8') + b'\n')
                            tmp_output_file.write(b'+\n')
                            tmp_output_file.write(qualities[j] + b'\n')
                            read_counter += 1
                    current_fastq = []

                    if line_counter/4 >= batch_size:
                        tmp_output_file.close()
                        line_counter = 0

                        if exception_queue is not None:
                            while True:
                                try:
                                    input_queue.put(tmp_output_file.name, True, 5)
                                    monitor.increment()
                                    submission_counter += 1
                                    break
                                except Full:
                                    pass
                                try:
                                    exc = exception_queue.get(block=False)
                                except Empty:
                                    pass
                                else:
                                    worker_pool.terminate()
                                    raise Exception(exc)
                        else:
                            input_queue.put(tmp_output_file.name, True, 5)
                            monitor.increment()
                            submission_counter += 1
                        tmp_output_file = tempfile.NamedTemporaryFile(suffix='.fastq', delete=False,
                                                                      dir=output_folder, mode='w+b')
        tmp_output_file.close()
        if line_counter > 0:
            input_queue.put(tmp_output_file.name, True)
            monitor.increment()
            submission_counter += 1
    except Exception:
        import sys
        stacktrace = "".join(traceback.format_exception(*sys.exc_info()))
        logger.error(stacktrace)
        exception_queue.put(stacktrace)
    finally:
        if tmp_output_file is not None and not tmp_output_file.closed:
            tmp_output_file.close()
    logger.debug("Submitted {} FASTQ chunks".format(submission_counter))
    monitor.set_submitting(False)


def _resubmissions_to_queue(resubmission_queue, output_folder, batch_size,
                            input_queue, monitor, step_size=5, min_size=25,
                            trim_front=False,
                            exception_queue=None, worker_pool=None):
    """
    Collect resubmission files, process the FASTQ, and resubmit to input queue.

    :param resubmission_queue: queue for resubmission files
    :param output_folder: output folder (tmp)
    :param batch_size: Number of reads submitted to a monitor
    :param input_queue: queue for input (batch) FASTQ files
    :param monitor: worker monitor
    :param step_size: Step size for read trimming
    :param min_size: Minimum read size
    :param trim_front: If True, trims reads from front instead of back
    :param exception_queue: queue for exceptions
    :param worker_pool: Pool of mapping workers "_iterative_mapping_worker"
    :return:
    """
    collected_resubmits = 0
    tmp_output_file = None
    read_counter = 0
    try:
        tmp_output_file = tempfile.NamedTemporaryFile(suffix='.fastq', delete=False,
                                                      dir=output_folder, mode='w+b')

        while monitor.is_submitting() or collected_resubmits < monitor.value():
            logger.debug("Status: {}/{}".format(collected_resubmits, monitor.value()))
            monitor.set_resubmitting(False)
            resubmission_file = resubmission_queue.get(block=True)
            monitor.set_resubmitting(True)
            logger.debug('Got resubmission file {}'.format(resubmission_file))

            collected_resubmits += 1
            if resubmission_file is None:
                continue

            with io.open(resubmission_file) as f:

                current_fastq = []
                for i, line in enumerate(f):
                    current_fastq.append(line.rstrip())

                    if i % 4 == 3:  # FASTQ record is complete
                        try:
                            new_fastq = _trim_read(current_fastq, step_size=step_size,
                                                   min_size=min_size, front=trim_front)
                        except ValueError:
                            continue

                        if len(new_fastq[1]) != len(current_fastq[1]):
                            for fastq_line in new_fastq:
                                new_line = fastq_line + '\n'
                                new_line = new_line.encode('utf-8')
                                tmp_output_file.write(new_line)
                            read_counter += 1
                        current_fastq = []

                        if read_counter >= batch_size:
                            logger.debug("Resubmitting because batch is full ({})".format(read_counter))
                            tmp_output_file.close()
                            read_counter = 0

                            if exception_queue is not None:
                                while True:
                                    try:
                                        input_queue.put(tmp_output_file.name, True, 5)
                                        monitor.increment()
                                        break
                                    except Full:
                                        pass
                                    try:
                                        exc = exception_queue.get(block=False)
                                    except Empty:
                                        pass
                                    else:
                                        worker_pool.terminate()
                                        raise Exception(exc)
                            else:
                                input_queue.put(tmp_output_file.name, True)
                                monitor.increment()
                            tmp_output_file = tempfile.NamedTemporaryFile(suffix='.fastq', delete=False,
                                                                          dir=output_folder, mode='w+b')
            os.remove(resubmission_file)

            if (read_counter > 0 and monitor.workers_idle()
                    and not monitor.is_submitting() and resubmission_queue.empty()):
                logger.debug("Resubmitting prematurely ({}) because workers are waiting for input".format(read_counter))
                tmp_output_file.close()
                input_queue.put(tmp_output_file.name, True)
                monitor.increment()
                tmp_output_file = tempfile.NamedTemporaryFile(suffix='.fastq', delete=False,
                                                              dir=output_folder, mode='w+b')
    except Exception:
        import sys
        stacktrace = "".join(traceback.format_exception(*sys.exc_info()))
        logger.error(stacktrace)
        exception_queue.put(stacktrace)
    finally:
        monitor.set_resubmitting(False)
        if tmp_output_file is not None and not tmp_output_file.closed:
            tmp_output_file.close()
        os.remove(tmp_output_file.name)


def iterative_mapping(fastq_file, sam_file, mapper, tmp_folder=None, threads=1, min_size=25, step_size=5,
                      batch_size=200000, trim_front=False, restriction_enzyme=None):
    """
    Iteratively map sequencing reads using the provided mapper.

    Will attempt to align a read using mapper. If unsuccessful, will
    truncate the read by step_size and attempt to align again. This is
    repeated until a successful alignment is found or the read gets
    truncated below min_size.

    :param fastq_file: An input FASTQ file path with reds to align
    :param sam_file: An output file path for sequencing results.
                     If it ends with '.bam' will compress output
                     in bam format.
    :param mapper: An instance of :class:`Mapper`, e.g. :class:`Bowtie2Mapper`.
                   Override :class:`Mapper` for creating your own custom mappers.
    :param tmp_folder: A temporary folder for outputting subsets of FASTQ files
    :param threads: Number of mapper threads to use in parallel.
    :param min_size: Minimum length of read for which an alignment is attempted.
    :param step_size: Number of base pairs by which to truncate read.
    :param batch_size: Maximum number of reads processed in one batch
    :param trim_front: Trim bases from front of read instead of back
    :param restriction_enzyme: If provided, will calculate the expected ligation
                               junction between reads and split reads accordingly.
                               Both ends will be attempted to map. Can be the name
                               of a restriction enzyme or a restriction pattern
                               (e.g. A^AGCT_T)
    """
    if tmp_folder is None:
        tmp_folder = tempfile.mkdtemp()
    else:
        tmp_folder = tempfile.mkdtemp(dir=tmp_folder)
    logger.debug('Tmp folder: {}'.format(tmp_folder))

    try:
        mp_manager = mp.Manager()
        input_queue = mp_manager.Queue(maxsize=threads)
        resubmission_queue = mp_manager.Queue()
        output_queue = mp_manager.Queue()
        exception_queue = mp_manager.Queue()
        monitor = Monitor(manager=mp_manager)

        monitor.set_submitting(True)

        worker_pool = mp.Pool(threads, _iterative_mapping_worker,
                              (mapper, input_queue, tmp_folder, output_queue,
                               resubmission_queue, monitor, exception_queue))

        t_resub = threading.Thread(target=_resubmissions_to_queue, args=(resubmission_queue, tmp_folder,
                                                                         batch_size, input_queue, monitor,
                                                                         step_size, min_size, trim_front,
                                                                         exception_queue, worker_pool))
        t_resub.daemon = True
        t_resub.start()

        t_sub = threading.Thread(target=_fastq_to_queue, args=(fastq_file, tmp_folder,
                                                               batch_size, input_queue,
                                                               monitor, exception_queue,
                                                               worker_pool, restriction_enzyme))
        t_sub.daemon = True
        t_sub.start()

        # check BAM status
        if not sam_file.endswith('bam'):
            intermediate_sam_file = sam_file
            convert_to_bam = False
            logger.info("Starting to output alignments to SAM file {}".format(sam_file))
        else:
            intermediate_sam_file = os.path.join(tmp_folder, 'intermediate.sam')
            convert_to_bam = True
            logger.info("Starting to output alignments to intermediate SAM file {}".format(intermediate_sam_file))

        ligation_name_pattern = re.compile('(.+)__(\d)+$')
        sam_counter = 0
        with open(intermediate_sam_file, 'w') as o:
            while (sam_counter < monitor.value() or monitor.is_resubmitting()
                   or monitor.is_submitting() or not monitor.workers_idle()):
                try:
                    exc = exception_queue.get(block=False)
                except Empty:
                    pass
                else:
                    worker_pool.terminate()
                    raise Exception(exc)

                try:
                    partial_sam_file = output_queue.get(block=True, timeout=10)
                    logger.debug('Processing output file {}'.format(partial_sam_file))

                    with open(partial_sam_file, 'r') as f:
                        for line in f:
                            if line.startswith('@'):
                                if sam_counter == 0:
                                    o.write(line)
                                continue

                            line = line.rstrip()
                            if line == '':
                                continue

                            fields = line.split("\t")
                            m = ligation_name_pattern.match(fields[0])
                            if m is not None:
                                fields[0] = m.group(1)
                                fields[-1] += '\tZL:i:{}'.format(m.group(2))

                            o.write("\t".join(fields) + '\n')

                    os.remove(partial_sam_file)

                    sam_counter += 1
                    logger.debug('Got {}/{} SAM files'.format(sam_counter, monitor.value()))
                except Empty:
                    pass

        t_sub.join()
        t_resub.join()

        if convert_to_bam:
            logger.info("Converting intermediate SAM file to BAM ({})".format(sam_file))
            try:
                pysam.view('-b', '-o', sam_file, intermediate_sam_file, catch_stdout=False)
                logger.info("Success, removing intermediate.")
                os.remove(intermediate_sam_file)
            except pysam.SamtoolsError:
                logger.error("Could not convert to BAM, but your "
                             "output is still in {}".format(intermediate_sam_file))
    finally:
        logger.debug(tmp_folder)
        shutil.rmtree(tmp_folder, ignore_errors=True)

    return sam_file



import tables

import genomic_regions as gr
from .helpers import vector_enrichment_profile
from ..matrix import RegionMatrixTable, Edge
import numpy as np
from ..tools.general import RareUpdateProgressBar
from ..config import config
from ..regions import Genome
from future.utils import string_types
from Bio.SeqUtils import GC as calculate_gc_content
import logging

logger = logging.getLogger(__name__)


class ABCompartmentMatrix(RegionMatrixTable):
    """
    Class representing O/E correlation matrix used to derive AB compartments.

    You can generate an :class:`~ABCompartmentMatrix` from a Hic object
    using the :func:`~ABCompartmentMatrix.from_hic` class method.

    .. code::

        hic = fanc.load("path/to/file.hic")
        ab = ABCompartmentMatrix.from_hic(hic)

    The :code:`ab` object can then be used to calculate compartmentalisation
    eigenvectors and A/B compartment assignments:

    .. code::

        ev = ab.eigenvector()
        domains = ab.domains()

    For more robust A and B calls, you can use a genome file (FASTA) to
    orient the eigenvector so that regions with higher GC content on average
    get assigned positive EV values:

    .. code::

        ev = ab.eigenvector(genome="path/to/genome.fa")
        domains = ab.domains(genome="path/to/genome.fa")

    Finally, you can calculate an AB compertment enrichment profile using

    .. code::

        profile, ev_cutoffs = ab.enrichment_profile(hic)

        # or with genome to orient the EV
        profile, ev_cutoffs = ab.enrichment_profile(hic, genome="path/to/genome.fa")

    """

    _classid = 'ABCOMPARTMENTMATRIX'

    def __init__(self, file_name=None, mode='r', tmpdir=None):
        RegionMatrixTable.__init__(self, file_name=file_name, mode=mode, tmpdir=tmpdir,
                                   additional_region_fields={
                                       'name': tables.StringCol(1, pos=0),
                                       'score': tables.Float32Col(pos=1),
                                   })
        if 'per_chromosome' not in self.meta:
            self.meta['per_chromosome'] = True
        if 'oe_per_chromosome' not in self.meta:
            self.meta['oe_per_chromosome'] = True
        if 'has_ev' not in self.meta:
            self.meta['has_ev'] = False
        if 'eigenvector' not in self.meta:
            self.meta['eigenvector'] = 0
        if 'gc' not in self.meta:
            self.meta['gc'] = False

    @classmethod
    def from_hic(cls, hic, file_name=None, tmpdir=None,
                 per_chromosome=True, oe_per_chromosome=None):
        ab_matrix = cls(file_name=file_name, mode='w', tmpdir=tmpdir)
        ab_matrix.add_regions(hic.regions, preserve_attributes=False)
        ab_matrix.meta.per_chromosome = per_chromosome
        ab_matrix.meta.oe_per_chromosome = oe_per_chromosome

        if per_chromosome:
            if oe_per_chromosome is None:
                oe_per_chromosome = True
            chromosomes = hic.chromosomes()
            with RareUpdateProgressBar(max_value=len(chromosomes), silent=config.hide_progressbars,
                                       prefix="AB") as pb:
                for chr_ix, chromosome in enumerate(chromosomes):
                    m = hic.matrix((chromosome, chromosome), oe=True, oe_per_chromosome=oe_per_chromosome)
                    with np.errstate(divide='ignore', invalid='ignore'):
                        corr_m = np.corrcoef(m)

                    logger.debug("Chromosome {}".format(chromosome))
                    for i, row_region in enumerate(m.row_regions):
                        for j, col_region in enumerate(m.col_regions):
                            if j < i:
                                continue

                            source = row_region.ix
                            sink = col_region.ix

                            try:
                                if np.isnan(corr_m[i, j]):
                                    continue
                                ab_matrix.add_edge(Edge(source=source, sink=sink, weight=corr_m[i, j]))
                            except IndexError:
                                pass
                    pb.update(chr_ix)
        else:
            if oe_per_chromosome is None:
                oe_per_chromosome = False
            m = hic.matrix(oe=True, oe_per_chromosome=oe_per_chromosome)
            corr_m = np.corrcoef(m)
            with RareUpdateProgressBar(max_value=m.shape[0], silent=config.hide_progressbars,
                                       prefix="AB") as pb:
                for i, row_region in enumerate(m.row_regions):
                    for j in range(i, len(m.col_regions)):
                        if j < i:
                            continue
                        col_region = m.row_regions[j]
                        source = row_region.ix
                        sink = col_region.ix
                        ab_matrix.add_edge(Edge(source=source, sink=sink, weight=corr_m[i, j]))
                    pb.update(i)

        ab_matrix.flush()

        mappable = hic.mappable()
        ab_matrix.region_data('valid', mappable)

        return ab_matrix

    def eigenvector(self, sub_region=None, genome=None, eigenvector=0,
                    per_chromosome=None, oe_per_chromosome=None,
                    exclude_chromosomes=None, force=False):
        if per_chromosome is None:
            per_chromosome = self.meta.per_chromosome

        if oe_per_chromosome is None:
            oe_per_chromosome = self.meta.oe_per_chromosome

        if exclude_chromosomes is None:
            exclude_chromosomes = set()
        else:
            exclude_chromosomes = set(exclude_chromosomes)

        if (not force and self.meta.has_ev and
                oe_per_chromosome == self.meta.oe_per_chromosome and
                per_chromosome == self.meta.per_chromosome and
                eigenvector == self.meta.eigenvector and
                (genome is not None) == self.meta.gc):
            logger.info("Returning pre-calculated eigenvector!")
            ev = np.array([region.score for region in self.regions])
        else:
            ev = np.zeros(len(self.regions))
            if per_chromosome:
                for chromosome_sub in self.chromosomes():
                    m = self.matrix((chromosome_sub, chromosome_sub))
                    m[np.isnan(m)] = 0
                    w, v = np.linalg.eig(m)
                    # v might end up being masked
                    if hasattr(v, 'mask'):
                        v.mask = False
                    ab_vector = v[:, eigenvector]
                    for i, region in enumerate(m.row_regions):
                        ev[region.ix] = ab_vector[i]
            else:
                m = self.matrix()
                m[np.isnan(m)] = 0
                w, v = np.linalg.eig(m)
                # v might end up being masked
                if hasattr(v, 'mask'):
                    v.mask = False
                ab_vector = v[:, eigenvector]
                for i, region in enumerate(m.row_regions):
                    ev[region.ix] = ab_vector[i]

            if genome is not None:
                logger.info("Using GC content to orient eigenvector...")
                close_genome = False
                if isinstance(genome, string_types):
                    genome = Genome.from_string(genome, mode='r')
                    close_genome = True

                gc_content = [np.nan] * len(self.regions)
                for chromosome_sub in self.chromosomes():
                    if chromosome_sub in exclude_chromosomes:
                        continue
                    logger.debug("{}".format(chromosome_sub))
                    chromosome_sequence = genome[chromosome_sub].sequence
                    for region in self.regions(chromosome_sub):
                        s = chromosome_sequence[region.start - 1:region.end]
                        gc_content[region.ix] = calculate_gc_content(s)
                gc_content = np.array(gc_content)

                if close_genome:
                    genome.close()

                # use gc content to orient AB domain vector per chromosome
                cb = self.chromosome_bins
                for chromosome_sub, (start, end) in cb.items():
                    ev_sub = ev[start:end]
                    gc_sub = gc_content[start:end]
                    a_ixs = np.where(ev_sub >= 0.)
                    b_ixs = np.where(ev_sub < 0.)
                    gc_a = np.nanmean(gc_sub[a_ixs])
                    gc_b = np.nanmean(gc_sub[b_ixs])

                    logger.debug("Chromosome {} GC content A: {}, B: {}".format(
                        chromosome_sub, gc_a, gc_b
                    ))

                    if gc_a < gc_b:  # AB compartments are reversed!
                        logger.debug("Reversing EV for chromosome {}".format(chromosome_sub))
                        ev[start:end] = -1 * ev_sub

            try:
                self.region_data('score', ev)
                self.meta.per_chromosome = per_chromosome
                self.meta.oe_per_chromosome = oe_per_chromosome
                self.meta.has_ev = True
                self.meta.eigenvector = eigenvector
                self.meta.gc = genome is not None
                compartments = ['A' if s >= 0 else 'B' for s in ev]
                self.region_data('name', compartments)
            except OSError:
                logger.warning("Cannot save eigenvector to ABCompartmentMatrix since the file "
                               "is opened in read-only mode.")

        if sub_region is None:
            return ev

        regions = list(self.regions(sub_region))
        return ev[regions[0].ix:regions[-1].ix + 1]

    def domains(self, *args, **kwargs):
        ev = self.eigenvector(*args, **kwargs)

        region_subset = kwargs.get('region', None)

        current_ev_scores = []
        domains = []
        current_domain = None
        for i, region in enumerate(self.regions(region_subset)):
            if current_domain is None:
                current_domain = gr.GenomicRegion(chromosome=region.chromosome,
                                                  start=region.start, end=region.end,
                                                  name='A' if ev[i] >= 0 else 'B')
                current_ev_scores = [ev[i]]
            else:
                if current_domain.chromosome == region.chromosome and \
                        (ev[i] < 0) == (current_ev_scores[0] < 0):
                    current_domain.end = region.end
                    current_ev_scores.append(ev[i])
                else:
                    current_domain.score = np.nanmean(current_ev_scores)
                    domains.append(current_domain)
                    current_domain = gr.GenomicRegion(chromosome=region.chromosome,
                                                      start=region.start, end=region.end,
                                                      name='A' if ev[i] >= 0 else 'B')
                    current_ev_scores = [ev[i]]
        if current_domain is not None:
            current_domain.score = np.nanmean(current_ev_scores)
            domains.append(current_domain)

        return gr.RegionWrapper(domains)

    def enrichment_profile(self, hic, percentiles=(20.0, 40.0, 60.0, 80.0, 100.0),
                           only_gc=False, symmetric_at=None,
                           exclude_chromosomes=(),
                           intra_chromosomal=True, inter_chromosomal=False,
                           eigenvector=None, collapse_identical_breakpoints=False,
                           *args, **kwargs):

        if eigenvector is not None:
            ev = np.array(eigenvector)
            if not len(ev) == len(hic.regions):
                raise ValueError("Eigenvector must have same number of "
                                 "entries ({}) as regions in genome ({})!"
                                 .format(len(ev), len(hic.regions)))
        else:
            logger.info("Generating profile...")
            if only_gc:
                genome = kwargs.get('genome', None)
                if genome is None:
                    raise ValueError("genome cannot be 'None' when using GC content "
                                     "for compartment analysis.")
                # calculate GC content
                if isinstance(genome, string_types):
                    logger.info("Loading genome...")
                    genome = Genome.from_string(genome)

                logger.info("Calculating GC content...")
                gc_content = [np.nan] * len(self.regions)
                for chromosome in self.chromosomes():
                    logger.info("{}".format(chromosome))
                    chromosome_sequence = genome[chromosome].sequence
                    for region in self.regions(chromosome):
                        s = chromosome_sequence[region.start - 1:region.end]
                        gc_content[region.ix] = calculate_gc_content(s)
                gc_content = np.array(gc_content)
                ev = gc_content
            else:
                ev = self.eigenvector(exclude_chromosomes=exclude_chromosomes, *args, **kwargs)

        mappable = self.mappable()
        return vector_enrichment_profile(hic, ev, mappable=mappable,
                                         per_chromosome=kwargs.get('per_chromosome', True),
                                         percentiles=percentiles, symmetric_at=symmetric_at,
                                         exclude_chromosomes=exclude_chromosomes,
                                         intra_chromosomal=intra_chromosomal,
                                         inter_chromosomal=inter_chromosomal,
                                         collapse_identical_breakpoints=collapse_identical_breakpoints)
